import pandas as pd

df = pd.read_pickle("game_player.pickle")
df = df.drop_duplicates(subset=["game_id", "player_id"])
gt = pd.read_pickle("game_team.pickle")
game = pd.read_pickle("game.pickle")
possessions = pd.read_pickle("possessions.pickle")
lineups = pd.read_pickle("lineups.pickle")

possessions = possessions.merge(
    lineups.rename(columns={"lineup": "lineup_offense"}),
    left_on=["lineup_id_offense"],
    right_on=["lineup_id"],
    how="left",
)
possessions = possessions.merge(
    lineups.rename(columns={"lineup": "lineup_defense"}),
    left_on=["lineup_id_defense"],
    right_on=["lineup_id"],
    how="left",
)

exploded = possessions.explode(["lineup_offense"]).rename(
    columns={"lineup_offense": "player_id"}
)
exploded["net_seconds"] = (
    exploded["seconds_played_end"] - exploded["seconds_played_start"]
)
grp = (
    exploded.groupby(["game_id", "player_id"])["points", "net_seconds"]
    .sum()
    .reset_index()
)


q2 = possessions[possessions["game_id"] == "0011900062"]

game = game.rename(columns={"minutes": "game_minutes"})
gt = gt.rename(columns={"pace": "team_pace", "possessions": "team_possessions"})
df = df.merge(game[["game_id", "game_minutes", "start_date"]], on="game_id", how="left")
df = df.merge(
    gt[
        [
            "game_id",
            "team_id",
            "team_id_opponent",
            "score",
            "score_opponent",
            "won",
            "location",
            "team_possessions",
            "team_pace",
        ]
    ],
    on=["game_id", "team_id"],
    how="inner",
)
df.loc[df["won"] == True, "won"] = 1
df.loc[df["won"] == False, "won"] = 0

# randomly select 10% of game_ids

df = df[
    [
        "team_id",
        "start_date",
        "game_id",
        "player_id",
        "player_name",
        "start_position",
        "team_id_opponent",
        "points",
        "game_minutes",
        "pace",
        "possessions",
        "minutes",
        "won",
        "plus_minus",
        "location",
        "score",
        "score_opponent",
        "team_possessions",
        "team_pace",
    ]
]
# df = df[df['start_date'].between('2022-10-17', '2023-02-01')]

df.to_pickle("game_player_full.pickle")



from abc import ABC, abstractmethod
from typing import Optional, Union

import pandas as pd

from player_performance_ratings.ratings.rating_calculators import MatchRatingGenerator
from player_performance_ratings.ratings.enums import (
    RatingKnownFeatures,
    RatingHistoricalFeatures,
)

from player_performance_ratings.data_structures import (
    Match,
    PlayerRating,
    TeamRating,
    ColumnNames,
)


class RatingGenerator(ABC):

    def __init__(
        self,
        performance_column: str,
        non_estimator_known_features_out: Optional[list[RatingKnownFeatures]],
        historical_features_out: Optional[list[RatingHistoricalFeatures]],
        match_rating_generator: MatchRatingGenerator,
        seperate_player_by_position: Optional[bool] = False,
        prefix: str = "",
    ):
        self.performance_column = performance_column
        self.seperate_player_by_position = seperate_player_by_position
        self.match_rating_generator = match_rating_generator
        self._known_features_out = []
        self._historical_features_out = historical_features_out or []
        self.non_estimator_known_features_out = non_estimator_known_features_out
        self.prefix = prefix
        self.column_names = None
        self._calculated_match_ids = []

    def reset_ratings(self):
        self._calculated_match_ids = []

    @abstractmethod
    def generate_historical_by_matches(
        self,
        matches: list[Match],
        column_names: ColumnNames,
        historical_features_out: Optional[list[RatingHistoricalFeatures]] = None,
        known_features_out: Optional[list[RatingKnownFeatures]] = None,
    ) -> dict[Union[RatingKnownFeatures, RatingHistoricalFeatures], list[float]]:
        pass

    @abstractmethod
    def generate_historical(
        self,
        df: pd.DataFrame,
        column_names: ColumnNames,
        historical_features_out: Optional[list[RatingHistoricalFeatures]] = None,
        known_features_out: Optional[list[RatingKnownFeatures]] = None,
    ) -> pd.DataFrame:
        pass

    @abstractmethod
    def generate_future(
        self,
        df: Optional[pd.DataFrame],
        matches: Optional[list[Match]] = None,
        historical_features_out: Optional[list[RatingHistoricalFeatures]] = None,
        known_features_out: Optional[list[RatingKnownFeatures]] = None,
    ) -> pd.DataFrame:
        pass

    @property
    def known_features_out(self) -> list[str]:
        """
        Rating features that do not contain leakage. Thus they can be passed into an estimator
        """
        return self._known_features_out

    @property
    def features_out(
        self,
    ) -> list[str]:
        """
        Contains both known and historical features
        """
        return [*self.known_features_out, *self.historical_features_out]

    @property
    def historical_features_out(self) -> list[str]:
        """
        Rating Features that contain leakge. Thus, they must not be passed into an estimator.
        They are only inteded to be used for data-analysis
        """
        if self._historical_features_out:
            return self._historical_features_out
        return []

    @property
    def known_features_return(self) -> list[str]:
        """
        Rating features that do not contain leakage.
        Contains both features intended to be passed to the estimator
         and other features that are not intended to be passed to the estimator
        """
        if self.non_estimator_known_features_out:
            return list(
                set(self.non_estimator_known_features_out + self.known_features_out)
            )
        return self.known_features_out

    @property
    def player_ratings(self) -> dict[str, PlayerRating]:
        return dict(
            sorted(
                self.match_rating_generator.player_ratings.items(),
                key=lambda item: item[1].rating_value,
                reverse=True,
            )
        )

    @property
    def team_ratings(self) -> list[TeamRating]:
        team_id_ratings: list[TeamRating] = []
        teams = self.match_rating_generator.teams
        player_ratings = self.player_ratings
        for id, team in teams.items():
            team_player_ratings = [player_ratings[p] for p in team.player_ids]
            team_rating_value = sum(
                [p.rating_value for p in team_player_ratings]
            ) / len(team_player_ratings)
            team_id_ratings.append(
                TeamRating(
                    id=team.id,
                    name=team.name,
                    players=team_player_ratings,
                    last_match_day_number=team.last_match_day_number,
                    rating_value=team_rating_value,
                )
            )

        return list(
            sorted(team_id_ratings, key=lambda team: team.rating_value, reverse=True)
        )

    @property
    def calculated_match_ids(self) -> list[str]:
        return self._calculated_match_ids

    def _validate_match(self, match: Match):
        if len(match.teams) < 2:
            raise ValueError(f"{match.id} only contains {len(match.teams)} teams")



import copy
from typing import Optional, List, Union

from player_performance_ratings import Pipeline, ColumnNames
from player_performance_ratings.ratings.performance_generator import (
    PerformancesGenerator,
)

from player_performance_ratings.predictor._base import BasePredictor
from player_performance_ratings.ratings.rating_generator import RatingGenerator
from player_performance_ratings.transformers.base_transformer import (
    BasePerformancesTransformer,
    BaseTransformer,
    BaseLagGenerator,
)


class PipelineFactory:

    def __init__(
        self,
        predictor: BasePredictor,
        column_names: ColumnNames,
        rating_generators: Optional[
            Union[RatingGenerator, list[RatingGenerator]]
        ] = None,
        performances_generator: Optional[PerformancesGenerator] = None,
        pre_lag_transformers: Optional[List[BaseTransformer]] = None,
        lag_generators: Optional[List[BaseLagGenerator]] = None,
        post_lag_transformers: Optional[List[BaseTransformer]] = None,
    ):
        self.post_lag_transformers = post_lag_transformers or []
        self.pre_lag_transformers = pre_lag_transformers or []
        self.lag_generators = lag_generators or []
        self.column_names = column_names
        self.rating_generators = rating_generators or []
        if isinstance(self.rating_generators, RatingGenerator):
            self.rating_generators = [self.rating_generators]

        self.predictor = predictor
        self.performances_generator = performances_generator

    def create(
        self,
        performances_generator: Optional[PerformancesGenerator] = None,
        rating_generators: Optional[list[RatingGenerator]] = None,
        lag_generators: Optional[List[BaseLagGenerator]] = None,
        pre_lag_transformers: Optional[List[BaseTransformer]] = None,
        post_lag_transformers: Optional[List[BaseTransformer]] = None,
        predictor: Optional[BasePredictor] = None,
    ) -> Pipeline:
        rating_generators = (
            rating_generators
            if rating_generators is not None
            else [copy.deepcopy(r) for r in self.rating_generators]
        )
        performances_generator = (
            performances_generator
            if performances_generator is not None
            else copy.deepcopy(self.performances_generator)
        )
        pre_lag_transformers = (
            pre_lag_transformers
            if pre_lag_transformers is not None
            else [copy.deepcopy(r.reset()) for r in self.pre_lag_transformers]
        )
        post_lag_transformers = (
            post_lag_transformers
            if post_lag_transformers is not None
            else [copy.deepcopy(r.reset()) for r in self.post_lag_transformers]
        )
        lag_generators = (
            lag_generators
            if lag_generators is not None
            else [copy.deepcopy(r.reset()) for r in self.lag_generators]
        )

        predictor = predictor if predictor is not None else self.predictor

        return Pipeline(
            rating_generators=rating_generators,
            performances_generator=performances_generator,
            lag_generators=copy.deepcopy(lag_generators),
            pre_lag_transformers=copy.deepcopy(pre_lag_transformers),
            post_lag_transformers=copy.deepcopy(post_lag_transformers),
            predictor=predictor,
            column_names=self.column_names,
        )



import copy
from typing import Optional
import polars as pl
import pandas as pd
from player_performance_ratings import ColumnNames

from player_performance_ratings.scorer.score import BaseScorer
from player_performance_ratings.transformers.base_transformer import (
    BaseTransformer,
    BaseLagGenerator,
)

from player_performance_ratings.cross_validator._base import CrossValidator
from player_performance_ratings.predictor._base import BasePredictor
from player_performance_ratings.utils import convert_pandas_to_polars


class MatchKFoldCrossValidator(CrossValidator):
    """
    Performs cross-validation by splitting the data into n_splits based on match_ids.
    """

    def __init__(
        self,
        match_id_column_name: str,
        date_column_name: str,
        scorer: Optional[BaseScorer] = None,
        min_validation_date: Optional[str] = None,
        n_splits: int = 3,
    ):
        """
        :param match_id_column_name: The column name of the match_id
        :param date_column_name: The column name of the date
        :param scorer: The scorer to use for measuring the accuracy of the predictions on the validation dataset
        :param min_validation_date: The minimum date for which the cross-validation should start
        :param n_splits: The number of splits to perform
        """
        super().__init__(scorer=scorer, min_validation_date=min_validation_date)
        self.match_id_column_name = match_id_column_name
        self.date_column_name = date_column_name
        self.n_splits = n_splits
        self.min_validation_date = min_validation_date

    def generate_validation_df(
        self,
        df: pd.DataFrame,
        predictor: BasePredictor,
        column_names: ColumnNames,
        estimator_features: Optional[list[str]] = None,
        pre_lag_transformers: Optional[list[BaseTransformer]] = None,
        lag_generators: Optional[list[BaseLagGenerator]] = None,
        post_lag_transformers: Optional[list[BaseTransformer]] = None,
        return_features: bool = False,
        add_train_prediction: bool = False,
    ) -> pd.DataFrame:
        """
        Generate predictions on validation dataset.
        Training is performed N times on previous match_ids and predictions are made on match_ids that take place in the future.
        :param df: The dataframe to generate the validation dataset from
        :param predictor: The predictor to use for generating the predictions
        :param column_names: The column names to use for the match_id, team_id, and player_id
        :param estimator_features: The features to use for the estimator. If passed in, it will override the estimator_features in the predictor
        :param pre_lag_transformers: The transformers to use before the lag generators
        :param lag_generators: The lag generators to use
        :param post_lag_transformers: The transformers to use after the lag generators
        :param return_features: Whether to return the features generated by the generators and the lags. If false it will only return the original columns and the predictions
        :param add_train_prediction: Whether to also calculate and return predictions for the training dataset.
            This can be beneficial for 2 purposes:
            1. To see how well the model is fitting the training data
            2. If the output of the predictions is used as input for another model

            If set to false it will only return the predictions for the validation dataset
        """

        predictor = copy.deepcopy(predictor)
        validation_dfs = []
        ori_cols = df.columns.tolist()

        estimator_features = estimator_features or []
        pre_lag_transformers = pre_lag_transformers or []
        lag_generators = lag_generators or []
        post_lag_transformers = post_lag_transformers or []

        if not self.min_validation_date:
            unique_dates = df[self.date_column_name].unique()
            median_number = len(unique_dates) // 2
            self.min_validation_date = unique_dates[median_number]

        df = df.assign(__cv_match_number=range(len(df)))
        min_validation_match_number = df[
            df[self.date_column_name] >= self.min_validation_date
        ]["__cv_match_number"].min()
        if not pre_lag_transformers:
            for lag_transformer in lag_generators:
                lag_transformer.reset()
                df = lag_transformer.generate_historical(df, column_names=column_names)

        max_match_number = df["__cv_match_number"].max()
        train_cut_off_match_number = min_validation_match_number
        step_matches = (max_match_number - min_validation_match_number) / self.n_splits
        train_df = df[(df["__cv_match_number"] < train_cut_off_match_number)]
        if len(train_df) < 0:
            raise ValueError(
                f"train_df is empty. train_cut_off_day_number: {train_cut_off_match_number}. Select a lower validation_match value."
            )
        validation_df = df[
            (df["__cv_match_number"] >= train_cut_off_match_number)
            & (df["__cv_match_number"] <= train_cut_off_match_number + step_matches)
        ]

        for idx in range(self.n_splits):
            if pre_lag_transformers:
                for pre_lag_transformer in pre_lag_transformers:
                    pre_lag_transformer.reset()
                    train_df = pre_lag_transformer.fit_transform(
                        train_df, column_names=column_names
                    )
                    validation_df = pre_lag_transformer.transform(validation_df)
                for lag_idx, lag_transformer in enumerate(lag_generators):
                    count_remaining_polars = [
                        l
                        for l in lag_generators[lag_idx:]
                        if "Polars" in l.__class__.__name__
                    ] + [
                        l
                        for l in post_lag_transformers
                        if "Polars" in l.__class__.__name__
                    ]

                    if isinstance(train_df, pd.DataFrame) and len(
                        count_remaining_polars
                    ) == len(lag_generators[lag_idx:]) + len(post_lag_transformers):
                        train_df = convert_pandas_to_polars(train_df)
                        validation_df = convert_pandas_to_polars(validation_df)

                    lag_transformer.reset()
                    train_df = lag_transformer.generate_historical(
                        train_df, column_names=column_names
                    )
                    validation_df = lag_transformer.generate_historical(
                        validation_df, column_names=column_names
                    )

            for post_lag_transformer in post_lag_transformers:
                post_lag_transformer.reset()
                train_df = post_lag_transformer.fit_transform(
                    train_df, column_names=column_names
                )
                validation_df = post_lag_transformer.transform(validation_df)

            if isinstance(train_df, pl.DataFrame):
                train_df = train_df.to_pandas()
            if isinstance(validation_df, pl.DataFrame):
                validation_df = validation_df.to_pandas()
            predictor.train(train_df, estimator_features=estimator_features)

            if idx == 0 and add_train_prediction:
                train_df = train_df[
                    [c for c in train_df.columns if c not in predictor.columns_added]
                ]
                train_df = predictor.add_prediction(train_df)
                train_df = train_df.assign(**{self.validation_column_name: 0})
                validation_dfs.append(train_df)

            validation_df = validation_df[
                [c for c in validation_df.columns if c not in predictor.columns_added]
            ]
            validation_df = predictor.add_prediction(validation_df)
            validation_df = validation_df.assign(**{self.validation_column_name: 1})
            validation_dfs.append(validation_df)

            train_cut_off_match_number = train_cut_off_match_number + step_matches
            train_df = df[(df["__cv_match_number"] < train_cut_off_match_number)]

            if idx == self.n_splits - 2:
                validation_df = df[
                    (df["__cv_match_number"] >= train_cut_off_match_number)
                ]
            else:
                validation_df = df[
                    (df["__cv_match_number"] >= train_cut_off_match_number)
                    & (
                        df["__cv_match_number"]
                        < train_cut_off_match_number + step_matches
                    )
                ]

        concat_validation_df = pd.concat(validation_dfs).drop(
            columns=["__cv_match_number"]
        )
        if not return_features:
            concat_validation_df = concat_validation_df[[
                *ori_cols, *predictor.columns_added, self.validation_column_name
            ]]

        return concat_validation_df.drop_duplicates(
            [column_names.match_id, column_names.team_id, column_names.player_id]
        )



import math
from dataclasses import dataclass
from abc import ABC, abstractmethod
from enum import Enum
from typing import Optional, Callable, Union, Any
from narwhals.typing import FrameT
import narwhals as nw
import narwhals
import numpy as np
import pandas as pd
from polars.polars import col

from player_performance_ratings.consts import PredictColumnNames

import polars as pl


class Operator(Enum):
    EQUALS = "=="
    NOT_EQUALS = "!="
    GREATER_THAN = ">"
    LESS_THAN = "<"
    GREATER_THAN_OR_EQUALS = ">="
    LESS_THAN_OR_EQUALS = "<="
    IN = "in"
    NOT_IN = "not in"


@dataclass
class Filter:
    column_name: str
    value: Union[Any, list[Any]]
    operator: Operator


def _apply_filters_pandas(df: pd.DataFrame, filters: list[Filter]) -> pd.DataFrame:
    for filter in filters:
        if filter.operator == Operator.EQUALS:
            df = df[df[filter.column_name] == filter.value]
        elif filter.operator == Operator.NOT_EQUALS:
            df = df[df[filter.column_name] != filter.value]
        elif filter.operator == Operator.GREATER_THAN:
            df = df[df[filter.column_name] > filter.value]
        elif filter.operator == Operator.LESS_THAN:
            df = df[df[filter.column_name] < filter.value]
        elif filter.operator == Operator.GREATER_THAN_OR_EQUALS:
            df = df[df[filter.column_name] >= filter.value]
        elif filter.operator == Operator.LESS_THAN_OR_EQUALS:
            df = df[df[filter.column_name] <= filter.value]
        elif filter.operator == Operator.IN:
            df = df[df[filter.column_name].isin(filter.value)]
        elif filter.operator == Operator.NOT_IN:
            df = df[~df[filter.column_name].isin(filter.value)]

    return df


def _apply_filters_polars(df: pl.DataFrame, filters: list[Filter]) -> pl.DataFrame:
    for filter in filters:
        if filter.operator == Operator.EQUALS:
            df = df.filter(col(filter.column_name) == filter.value)
        elif filter.operator == Operator.NOT_EQUALS:
            df = df.filter(col(filter.column_name) != filter.value)
        elif filter.operator == Operator.GREATER_THAN:
            df = df.filter(col(filter.column_name) > filter.value)
        elif filter.operator == Operator.LESS_THAN:
            df = df.filter(col(filter.column_name) < filter.value)
        elif filter.operator == Operator.GREATER_THAN_OR_EQUALS:
            df = df.filter(col(filter.column_name) >= filter.value)
        elif filter.operator == Operator.LESS_THAN_OR_EQUALS:
            df = df.filter(col(filter.column_name) <= filter.value)
        elif filter.operator == Operator.IN:
            df = df.filter(col(filter.column_name).isin(filter.value))
        elif filter.operator == Operator.NOT_IN:
            df = df.filter(~col(filter.column_name).isin(filter.value))

    return df

@narwhals.narwhalify()
def apply_filters(df: FrameT, filters: list[Filter]) -> Union[pd.DataFrame, pl.DataFrame]:
    for filter in filters:
        if filter.operator == Operator.EQUALS:
            df = df.filter(nw.col(filter.column_name) == filter.value)
        elif filter.operator == Operator.NOT_EQUALS:
            df = df.filter(nw.col(filter.column_name) != filter.value)
        elif filter.operator == Operator.GREATER_THAN:
            df = df.filter(nw.col(filter.column_name) > filter.value)
        elif filter.operator == Operator.LESS_THAN:
            df = df.filter(nw.col(filter.column_name) < filter.value)
        elif filter.operator == Operator.GREATER_THAN_OR_EQUALS:
            df = df.filter(nw.col(filter.column_name) >= filter.value)
        elif filter.operator == Operator.LESS_THAN_OR_EQUALS:
            df = df.filter(nw.col(filter.column_name) <= filter.value)
        elif filter.operator == Operator.IN:
            df = df.filter(nw.col(filter.column_name).is_in(filter.value))
        elif filter.operator == Operator.NOT_IN:
            df = df.filter(~nw.col(filter.column_name).is_in(filter.value))

    return df.to_native()

class BaseScorer(ABC):

    def __init__(
            self,
            target: str,
            pred_column: str,
            validation_column: Optional[str],
            filters: Optional[list[Filter]] = None,
            granularity: Optional[list[str]] = None,
    ):
        self.target = target
        self.pred_column = pred_column
        self.validation_column = validation_column
        self.filters = []
        if validation_column:
            self.filters.append(Filter(column_name=self.validation_column, value=1, operator=Operator.EQUALS))
        self.granularity = granularity

    @abstractmethod
    def score(self, df: Union[pl.DataFrame, pd.DataFrame]) -> float:
        pass


class SklearnScorer(BaseScorer):

    def __init__(
            self,
            pred_column: str,
            scorer_function: Callable,
            target: Optional[str] = PredictColumnNames.TARGET,
            validation_column: Optional[str] = None,
            granularity: Optional[list[str]] = None,
            filters: Optional[list[Filter]] = None,
    ):
        """
        :param pred_column: The column name of the predictions
        :param scorer_function: SKlearn scorer function, e.g. los_loss
        :param target: The column name of the target
        :param validation_column: The column name of the validation column.
            If set, the scorer will be calculated only once the values of the validation column are equal to 1
        :param granularity: The columns to group by before calculating the score
        :param filters: The filters to apply before calculating
        """

        self.pred_column_name = pred_column
        self.scorer_function = scorer_function
        super().__init__(
            target=target,
            pred_column=pred_column,
            granularity=granularity,
            filters=filters,
            validation_column=validation_column,
        )

    @narwhals.narwhalify
    def score(self, df: Union[pd.DataFrame, pl.DataFrame]) -> float:

        df = apply_filters(df=df, filters=self.filters)
        if self.granularity:
            grouped = (
                df.groupby(self.granularity)
                .agg([
                    pl.col(self.pred_column_name).mean().alias(self.pred_column_name),
                    pl.col(self.target).mean().alias(self.target)
                ])
            )
        else:
            grouped = df

        if isinstance(grouped[self.pred_column_name].to_list()[0], list):
            return self.scorer_function(
                grouped[self.target],
                [item for item in grouped[self.pred_column_name].to_list()],
            )

        return self.scorer_function(
            grouped[self.target].to_list(),
            grouped[self.pred_column_name].to_list()
        )


class ProbabilisticMeanBias(BaseScorer):

    def __init__(
            self,
            pred_column: str,
            class_column_name: str = "classes",
            target: Optional[str] = PredictColumnNames.TARGET,
            validation_column: Optional[str] = None,
            granularity: Optional[list[str]] = None,
            filters: Optional[list[Filter]] = None,
    ):

        self.pred_column_name = pred_column
        self.class_column_name = class_column_name
        super().__init__(
            target=target,
            pred_column=pred_column,
            filters=filters,
            granularity=granularity,
            validation_column=validation_column,
        )

    def score(self, df: pd.DataFrame) -> float:
        df = df.copy()
        df = apply_filters(df, self.filters)
        df.reset_index(drop=True, inplace=True)

        distinct_classes_variations = df.drop_duplicates(subset=[self.class_column_name])[
            self.class_column_name
        ].tolist()

        sum_lrs = [0 for _ in range(len(distinct_classes_variations))]
        sum_lr = 0
        for variation_idx, distinct_class_variation in enumerate(
                distinct_classes_variations
        ):

            if not isinstance(distinct_class_variation, list):
                if math.isnan(distinct_class_variation):
                    continue

            rows_target_group = df[
                df[self.class_column_name].apply(lambda x: x == distinct_class_variation)
            ]
            probs = rows_target_group[self.pred_column_name]
            last_column_name = f"prob_under_{distinct_class_variation[0] - 0.5}"
            rows_target_group[last_column_name] = probs.apply(lambda x: x[0])

            for idx, class_ in enumerate(distinct_class_variation[1:]):

                prob_under = "prob_under_" + str(class_ + 0.5)
                rows_target_group[prob_under] = (
                        probs.apply(lambda x: x[idx + 1])
                        + rows_target_group[last_column_name]
                )

                count_exact = len(
                    rows_target_group[rows_target_group["__target"] == class_]
                )
                weight_class = count_exact / len(rows_target_group)

                if self.granularity:
                    grouped = (
                        rows_target_group.groupby(self.granularity + ["__target"])[prob_under]
                        .mean()
                        .reset_index()
                    )
                else:
                    grouped = rows_target_group

                grouped["min"] = 0.0001
                grouped["max"] = 0.9999
                grouped[prob_under] = np.minimum(grouped["max"], grouped[prob_under])
                grouped[prob_under] = np.maximum(grouped["min"], grouped[prob_under])

                grouped.loc[grouped['__target'] <= class_, "__went_under"] = 1
                grouped.loc[grouped['__target'] > class_, "__went_under"] = 0

                under_prob_mean = grouped[prob_under].mean()
                under_actual_mean = grouped['__went_under'].mean()

                overbias = under_prob_mean - under_actual_mean
                sum_lrs[variation_idx] += overbias * weight_class

                last_column_name = prob_under
            sum_lr += sum_lrs[variation_idx] * len(rows_target_group) / len(df)
        return sum_lr


class OrdinalLossScorerPolars(BaseScorer):
    """TODO: Support different types of probability-class constructions"""

    def __init__(
            self,
            pred_column: str,
            target: Optional[str] = PredictColumnNames.TARGET,
            validation_column: Optional[str] = None,
            granularity: Optional[list[str]] = None,
            filters: Optional[list[Filter]] = None,
    ):
        self.pred_column_name = pred_column
        self.granularity = granularity
        super().__init__(
            target=target,
            pred_column=pred_column,
            filters=filters,
            granularity=granularity,
            validation_column=validation_column,
        )

    def score(self, df: pl.DataFrame) -> float:
        df = apply_filters(df, self.filters)
        field_names = [int(field) for field in df[self.pred_column].struct.fields]
        min_field = min(field_names)
        df = df.with_columns(
            [pl.col(self.pred_column).struct.field(str(field)).alias(f"prob_{field}") for field in field_names])
        prob_col_under = 'sum_prob_under'
        df = df.with_columns(pl.col(f'prob_{min_field}').alias(prob_col_under))
        sum_lr = 0
        count_higher_rows = len(df.filter(pl.col(self.target) > min_field))
        for class_ in field_names[1:]:

            count_exact = len(df.filter(pl.col(self.target) == class_-1))
            weight_class = count_exact / count_higher_rows

            df = df.with_columns([
                pl.lit(0.0001).alias("min"),
                pl.lit(0.9999).alias("max")
            ])

            df = df.with_columns(
                pl.max_horizontal(prob_col_under, "min").alias(prob_col_under),
                          )
            df = df.with_columns(
                pl.min_horizontal(prob_col_under, "max").alias(prob_col_under),
            )

            df = df.with_columns([
                pl.when(pl.col("__target") < class_)
                .then(pl.col(prob_col_under).log())
                .otherwise((1 - pl.col(prob_col_under)).log())
                .alias("log_loss")
            ])

            log_loss = df["log_loss"].mean()
            sum_lr -= log_loss * weight_class

            df = df.with_columns(
                (pl.col(f"prob_{class_}") + pl.col(prob_col_under)).alias(prob_col_under)
            )

        return sum_lr

class OrdinalLossScorer(BaseScorer):

    def __init__(
            self,
            pred_column: str,
            class_column_name: str = "classes",
            target: Optional[str] = PredictColumnNames.TARGET,
            validation_column: Optional[str] = None,
            granularity: Optional[list[str]] = None,
            filters: Optional[list[Filter]] = None,
    ):
        """
        :param pred_column: The column name of the predictions
        :param class_column_name: The column name that contains the integer values of the classes.
        :param target: The column name of the target
        :param validation_column: The column name of the validation column.
            If set, the scorer will be calculated only once the values of the validation column are equal to 1
        :param granularity: The columns to group by before calculating the score
        :param filters: The filters to apply before calculating
        """

        self.pred_column_name = pred_column
        self.class_column_name = class_column_name
        self.granularity = granularity
        super().__init__(
            target=target,
            pred_column=pred_column,
            filters=filters,
            granularity=granularity,
            validation_column=validation_column,
        )

    def score(self, df: pd.DataFrame) -> float:

        df = df.copy()
        df = apply_filters(df, self.filters)
        df.reset_index(drop=True, inplace=True)

        distinct_classes_variations = df.drop_duplicates(subset=[self.class_column_name])[
            self.class_column_name
        ].tolist()

        sum_lrs = [0 for _ in range(len(distinct_classes_variations))]
        sum_lr = 0
        for variation_idx, distinct_class_variation in enumerate(
                distinct_classes_variations
        ):

            if not isinstance(distinct_class_variation, list):
                if math.isnan(distinct_class_variation):
                    continue

            rows_target_group = df[
                df[self.class_column_name].apply(lambda x: x == distinct_class_variation)
            ]
            probs = rows_target_group[self.pred_column_name]
            last_column_name = f"prob_under_{distinct_class_variation[0] - 0.5}"
            rows_target_group[last_column_name] = probs.apply(lambda x: x[0])

            for idx, class_ in enumerate(distinct_class_variation[1:]):

                p_c = "prob_under_" + str(class_ + 0.5)
                rows_target_group[p_c] = (
                        probs.apply(lambda x: x[idx + 1])
                        + rows_target_group[last_column_name]
                )

                count_exact = len(
                    rows_target_group[rows_target_group["__target"] == class_]
                )
                weight_class = count_exact / len(rows_target_group)

                if self.granularity:
                    grouped = (
                        rows_target_group.groupby(self.granularity + ["__target"])[p_c]
                        .mean()
                        .reset_index()
                    )
                else:
                    grouped = rows_target_group

                grouped["min"] = 0.0001
                grouped["max"] = 0.9999
                grouped[p_c] = np.minimum(grouped["max"], grouped[p_c])
                grouped[p_c] = np.maximum(grouped["min"], grouped[p_c])
                grouped["log_loss"] = 0
                grouped.loc[grouped["__target"] <= class_, "log_loss"] = np.log(
                    grouped[p_c]
                )
                grouped.loc[grouped["__target"] > class_, "log_loss"] = np.log(
                    1 - grouped[p_c]
                )
                log_loss = grouped["log_loss"].mean()
                sum_lrs[variation_idx] -= log_loss * weight_class

                last_column_name = p_c
            sum_lr += sum_lrs[variation_idx] * len(rows_target_group) / len(df)
        return sum_lr



import logging
from typing import Optional, List

import numpy as np
import pandas as pd
from lightgbm import LGBMRegressor

from player_performance_ratings.predictor import Predictor
from player_performance_ratings.transformers.base_transformer import (
    BasePerformancesTransformer,
)


class SklearnEstimatorImputer(BasePerformancesTransformer):

    def __init__(
        self,
        features: list[str],
        target_name: str,
        estimator: Optional[LGBMRegressor] = None,
    ):
        super().__init__(features=features)
        self.estimator = estimator or LGBMRegressor()
        self.target_name = target_name

    def fit_transform(self, df: pd.DataFrame) -> pd.DataFrame:
        self.estimator.fit(df[self.features], df[self.target_name])
        df = df.assign(
            **{
                f"imputed_col_{self.target_name}": self.estimator.predict(
                    df[self.features]
                )
            }
        )
        df[self.target_name] = df[self.target_name].fillna(
            df[f"imputed_col_{self.target_name}"]
        )
        return df.drop(columns=[f"imputed_col_{self.target_name}"])

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df.assign(
            **{
                f"imputed_col_{self.target_name}": self.estimator.predict(
                    df[self.features]
                )
            }
        )
        df[self.target_name] = df[self.target_name].fillna(
            df[f"imputed_col_{self.target_name}"]
        )
        return df.drop(columns=[f"imputed_col_{self.target_name}"])

    @property
    def features_out(self) -> list[str]:
        return [self.target_name]


class PartialStandardScaler(BasePerformancesTransformer):

    def __init__(
        self,
        features: list[str],
        ratio: float = 0.55,
        target_mean: float = 0.5,
        max_value: float = 2,
        prefix: str = "",
    ):
        super().__init__(features=features)
        self.ratio = ratio
        self.target_mean = target_mean
        self.prefix = prefix
        self.max_value = max_value

        self._features_mean = {}
        self._features_std = {}
        self._features_out = []
        for feature in self.features:
            self._features_out.append(self.prefix + feature)

    def fit_transform(self, df: pd.DataFrame) -> pd.DataFrame:
        for feature in self.features:
            self._features_mean[feature] = (
                df[feature].replace([np.inf, -np.inf], np.nan).mean()
            )
            self._features_std[feature] = (
                df[feature].replace([np.inf, -np.inf], np.nan).std()
            )

        return self.transform(df)

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        for feature in self.features:
            new_feature = self.prefix + feature
            df = df.assign(
                **{
                    new_feature: (
                        (df[feature] - self._features_mean[feature])
                        / self._features_std[feature]
                    )
                    * self.ratio
                    + self.target_mean
                }
            )
            df[new_feature] = df[new_feature].clip(
                -self.max_value + self.target_mean, self.max_value
            )
        return df

    @property
    def features_out(self) -> list[str]:
        return self._features_out


class MinMaxTransformer(BasePerformancesTransformer):

    def __init__(
        self,
        features: list[str],
        quantile: float = 0.98,
        multiply_align: bool = False,
        add_align: bool = True,
        prefix: str = "",
    ):
        super().__init__(features=features)
        self.quantile = quantile
        self.prefix = prefix
        self._trained_mean_values = {}
        self._min_values = {}
        self._max_values = {}
        self.multiply_align = multiply_align
        self.add_align = add_align

        if self.quantile < 0 or self.quantile > 1:
            raise ValueError("quantile must be between 0 and 1")

        self._features_out = []

    def fit_transform(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df.copy()
        for feature in self.features:
            self._min_values[feature] = df[feature].quantile(1 - self.quantile)
            self._max_values[feature] = df[feature].quantile(self.quantile)

            if self._min_values[feature] == self._max_values[feature]:
                raise ValueError(
                    f"Feature {feature} has the same min and max value. This can lead to division by zero. "
                    f"This feature is not suited for MinMaxTransformer"
                )

            df[self.prefix + feature] = (df[feature] - self._min_values[feature]) / (
                self._max_values[feature] - self._min_values[feature]
            )
            df[self.prefix + feature].clip(0, 1, inplace=True)
            self._trained_mean_values[feature] = df[self.prefix + feature].mean()
            if self.multiply_align:
                df[self.prefix + feature] = (
                    df[self.prefix + feature] * 0.5 / self._trained_mean_values[feature]
                )
            if self.add_align:
                df[self.prefix + feature] = (
                    df[self.prefix + feature] + 0.5 - self._trained_mean_values[feature]
                )

            self._features_out.append(self.prefix + feature)

        return df

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df.copy()
        for feature in self.features:
            df[self.prefix + feature] = (df[feature] - self._min_values[feature]) / (
                self._max_values[feature] - self._min_values[feature]
            )
            df[self.prefix + feature].clip(0, 1, inplace=True)
            if self.multiply_align:
                df[self.prefix + feature] = (
                    df[self.prefix + feature] * 0.5 / self._trained_mean_values[feature]
                )
            if self.add_align:
                df[self.prefix + feature] = (
                    df[self.prefix + feature] + 0.5 - self._trained_mean_values[feature]
                )

        return df

    @property
    def features_out(self) -> list[str]:
        return self._features_out


class DiminishingValueTransformer(BasePerformancesTransformer):

    def __init__(
        self,
        features: List[str],
        cutoff_value: Optional[float] = None,
        quantile_cutoff: float = 0.93,
        excessive_multiplier: float = 0.8,
        reverse: bool = False,
    ):
        super().__init__(features=features)
        self.cutoff_value = cutoff_value
        self.excessive_multiplier = excessive_multiplier
        self.quantile_cutoff = quantile_cutoff
        self.reverse = reverse
        self._feature_cutoff_value = {}

    def fit_transform(self, df: pd.DataFrame) -> pd.DataFrame:

        for feature_name in self.features:
            if self.cutoff_value is None:
                if self.reverse:
                    self._feature_cutoff_value[feature_name] = df[
                        feature_name
                    ].quantile(1 - self.quantile_cutoff)
                else:
                    self._feature_cutoff_value[feature_name] = df[
                        feature_name
                    ].quantile(self.quantile_cutoff)
            else:
                self._feature_cutoff_value[feature_name] = self.cutoff_value

        return self.transform(df)

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        for feature_name in self.features:
            cutoff_value = self._feature_cutoff_value[feature_name]
            if self.reverse:
                df = df.assign(
                    **{
                        feature_name: lambda x: np.where(
                            x[feature_name] <= cutoff_value,
                            -(cutoff_value - df[feature_name])
                            * self.excessive_multiplier
                            + cutoff_value,
                            x[feature_name],
                        )
                    }
                )
            else:

                df = df.assign(
                    **{
                        feature_name: lambda x: np.where(
                            x[feature_name] >= cutoff_value,
                            (x[feature_name] - cutoff_value).clip(lower=0)
                            * self.excessive_multiplier
                            + cutoff_value,
                            x[feature_name],
                        )
                    }
                )

            df = df.assign(**{feature_name: df[feature_name].fillna(df[feature_name])})
        return df

    @property
    def features_out(self) -> list[str]:
        return self.features


class SymmetricDistributionTransformer(BasePerformancesTransformer):

    def __init__(
        self,
        features: List[str],
        granularity: Optional[list[str]] = None,
        skewness_allowed: float = 0.15,
        max_iterations: int = 50,
        min_excessive_multiplier: float = 0.04,
        prefix: str = "symmetric_",
    ):
        super().__init__(features=features)
        self.granularity = granularity
        self.skewness_allowed = skewness_allowed
        self.max_iterations = max_iterations
        self.min_excessive_multiplier = min_excessive_multiplier
        self.prefix = prefix

        self._diminishing_value_transformer = {}

    def fit_transform(self, df: pd.DataFrame) -> pd.DataFrame:

        if self.granularity:
            df = df.assign(
                __concat_granularity=df[self.granularity]
                .astype(str)
                .agg("_".join, axis=1)
            )

        for feature in self.features:
            if df[feature].min() == df[feature].max():
                raise ValueError(
                    f"SymmetricDistributionTransformer: {feature} has the same min and max value."
                )
            self._diminishing_value_transformer[feature] = {}
            if self.granularity:

                unique_values = df["__concat_granularity"].unique()
                for unique_value in unique_values:
                    rows = df[df["__concat_granularity"] == unique_value]

                    self._fit(
                        rows=rows, feature=feature, granularity_value=unique_value
                    )

            else:
                self._fit(rows=df, feature=feature, granularity_value=None)

        return self.transform(df)

    def _fit(
        self, rows: pd.DataFrame, feature: str, granularity_value: Optional[str]
    ) -> None:

        skewness = rows[feature].skew()
        excessive_multiplier = 0.8
        quantile_cutoff = 0.95

        iteration = 0
        while (
            abs(skewness) > self.skewness_allowed
            and len(rows) > 10
            and iteration < self.max_iterations
        ):

            if skewness < 0:
                reverse = True
            else:
                reverse = False

            self._diminishing_value_transformer[feature][granularity_value] = (
                DiminishingValueTransformer(
                    features=[feature],
                    reverse=reverse,
                    excessive_multiplier=excessive_multiplier,
                    quantile_cutoff=quantile_cutoff,
                )
            )
            transformed_rows = self._diminishing_value_transformer[feature][
                granularity_value
            ].fit_transform(rows)
            new_excessive_multiplier = excessive_multiplier * 0.94
            if new_excessive_multiplier < self.min_excessive_multiplier:
                break
            excessive_multiplier = new_excessive_multiplier
            next_quantile_cutoff = quantile_cutoff * 0.994
            if (
                transformed_rows[feature].quantile(next_quantile_cutoff)
                > transformed_rows[feature].min()
            ):
                quantile_cutoff = next_quantile_cutoff
            iteration += 1
            skewness = transformed_rows[feature].skew()

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df.copy()
        if self.granularity:
            df = df.assign(
                __concat_granularity=df[self.granularity].apply(
                    lambda x: "_".join(x), axis=1
                )
            )

        for feature in self.features:
            out_feature = self.prefix + feature
            if self.granularity:

                unique_values = df["__concat_granularity"].unique()
                if len(unique_values) > 100:
                    logging.warning(
                        f"SymmetricDistributionTransformer: {feature} has more than 100 unique values."
                        f" This can lead to long runtimes. Consider setting a lower granularity"
                    )

                if len(self._diminishing_value_transformer[feature]) == 0:
                    df[out_feature] = df[feature]
                else:

                    for unique_value in unique_values:
                        rows = df[df["__concat_granularity"] == unique_value]

                        if unique_value in self._diminishing_value_transformer[feature]:
                            rows = self._diminishing_value_transformer[feature][
                                unique_value
                            ].transform(rows)

                        df.loc[
                            df["__concat_granularity"] == unique_value, out_feature
                        ] = rows[feature]

            else:
                if None in self._diminishing_value_transformer[feature]:
                    df = self._diminishing_value_transformer[feature][None].transform(
                        df
                    )

        if "__concat_granularity" in df.columns:
            df = df.drop(columns=["__concat_granularity"])
        return df

    @property
    def features_out(self) -> list[str]:
        return [self.prefix + feature for feature in self.features]


class GroupByTransformer(BasePerformancesTransformer):

    def __init__(
        self,
        features: list[str],
        granularity: list[str],
        agg_func: str = "mean",
        prefix: str = "mean_",
    ):
        super().__init__(features=features)
        self.granularity = granularity
        self.agg_func = agg_func
        self.prefix = prefix
        self._features_out = []
        self._grouped = None
        for feature in self.features:
            self._features_out.append(self.prefix + feature)

    def fit_transform(self, df: pd.DataFrame) -> pd.DataFrame:
        self._grouped = (
            df.groupby(self.granularity)[self.features]
            .agg(self.agg_func)
            .reset_index()
            .rename(
                columns={feature: self.prefix + feature for feature in self.features}
            )
        )

        return self.transform(df=df)

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        return df.merge(self._grouped, on=self.granularity, how="left")

    @property
    def features_out(self) -> list[str]:
        return self._features_out



from typing import Optional, Any, Union

import numpy as np
import pandas as pd

from player_performance_ratings.ratings import convert_df_to_matches
from player_performance_ratings.ratings.rating_calculators import (
    RatingMeanPerformancePredictor,
)
from player_performance_ratings.ratings.rating_calculators.performance_predictor import (
    RatingNonOpponentPerformancePredictor,
)
from player_performance_ratings.ratings.rating_calculators.match_rating_generator import (
    MatchRatingGenerator,
)
from player_performance_ratings.ratings.enums import (
    RatingKnownFeatures,
    RatingHistoricalFeatures,
)

from player_performance_ratings.data_structures import (
    Match,
    PreMatchRating,
    PreMatchTeamRating,
    ColumnNames,
    TeamRatingChange,
)
from player_performance_ratings.ratings.rating_generator import RatingGenerator


class UpdateRatingGenerator(RatingGenerator):
    """
    Generates ratings for players and teams based on the match-performance of the player and the ratings of the players and teams.
    Ratings are updated after a match is finished
    """

    def __init__(
        self,
        performance_column: str = "performance",
        match_rating_generator: Optional[MatchRatingGenerator] = None,
        known_features_out: Optional[list[RatingKnownFeatures]] = None,
        historical_features_out: Optional[list[RatingHistoricalFeatures]] = None,
        non_estimator_known_features_out: Optional[list[RatingKnownFeatures]] = None,
        distinct_positions: Optional[list[str]] = None,
        seperate_player_by_position: bool = False,
        prefix: str = "",
    ):
        """
        :param performance_column: The ratings will be updated by on the value of the column
        :param match_rating_generator: Passing in the MatchRatingGenerator allows for customisation of the classes and parameters used within it.
        :param known_features_out: A list of features where the information is available before the match is started.
            The pre-game ratings are an example of this whereas player-rating-change is a historical feature as it's first known after the match is finished.
            If none, default logic is generated based on the performance predictor used in the match_rating_generator
        :param historical_features_out: The types of historical rating-features the rating-generator should return.
            The historical_features cannot be used as estimator features as they contain leakage, however can be interesting for exploration
        :param non_estimator_known_features_out: Rating Features to return but which are not intended to be used as estimator_features for the predictor
        :param distinct_positions: If true, the rating_difference for each player relative to the opponent player by the same position will be generated and returned.
        :param seperate_player_by_position: Creates a unique identifier for each player based on the player_id and position.
            Set to true if a players skill-level is dependent on the position they play.
        """

        super().__init__(
            non_estimator_known_features_out=non_estimator_known_features_out,
            historical_features_out=historical_features_out,
            performance_column=performance_column,
            seperate_player_by_position=seperate_player_by_position,
            match_rating_generator=match_rating_generator or MatchRatingGenerator(),
            prefix=prefix,
        )
        self.distinct_positions = distinct_positions
        self._non_estimator_rating_features_out = non_estimator_known_features_out or []

        self._known_features_out = (
            [self.prefix + f for f in known_features_out]
            if known_features_out is not None
            else (
                [self.prefix + RatingKnownFeatures.RATING_MEAN_PROJECTED]
                if isinstance(
                    self.match_rating_generator.performance_predictor,
                    RatingMeanPerformancePredictor,
                )
                else (
                    [self.prefix + RatingKnownFeatures.PLAYER_RATING]
                    if isinstance(
                        self.match_rating_generator.performance_predictor,
                        RatingNonOpponentPerformancePredictor,
                    )
                    else [self.prefix + RatingKnownFeatures.RATING_DIFFERENCE_PROJECTED]
                )
            )
        )

        if self.distinct_positions:
            self._known_features_out += [
                self.prefix + RatingKnownFeatures.RATING_DIFFERENCE_POSITION + "_" + p
                for p in self.distinct_positions
            ]

    def reset_ratings(self):
        self._calculated_match_ids = []
        self.match_rating_generator.player_ratings = {}
        self.match_rating_generator._teams = {}
        self.match_rating_generator.start_rating_generator.reset()
        self.match_rating_generator.performance_predictor.reset()

    def generate_historical_by_matches(
        self,
        matches: list[Match],
        column_names: ColumnNames,
        historical_features_out: Optional[list[RatingHistoricalFeatures]] = None,
        known_features_out: Optional[list[RatingKnownFeatures]] = None,
    ) -> dict[Union[RatingKnownFeatures, RatingHistoricalFeatures], list[float]]:
        """
        Generate ratings by iterating over each match, calculate predicted performance and update ratings after the match is finished.

        :param matches: A list of Matches containing the data required to calculate ratings
        :param column_names: The column names of the dataframe. Only needed if you want to store the ratings in the class object in which case the df must also be passed.
        :param historical_features_out: The types of historical rating-features the rating-generator should return.

        :return: A dictionary containing a list of match-rating
        """
        self.column_names = column_names
        historical_features_out = (
            historical_features_out or self._historical_features_out
        )
        known_features_out = known_features_out or self._known_features_out
        potential_feature_values = self._generate_potential_feature_values(
            matches=matches
        )
        return {
            f: potential_feature_values[f]
            for f in list(set(known_features_out + historical_features_out + self._non_estimator_rating_features_out))
        }

    def generate_historical(
        self,
        df: pd.DataFrame,
        column_names: ColumnNames,
        historical_features_out: Optional[list[RatingHistoricalFeatures]] = None,
        known_features_out: Optional[list[RatingKnownFeatures]] = None,
    ) -> pd.DataFrame:
        """
        Generate ratings by iterating over the dataframe, calculate predicted performance and update ratings after the match is finished.

        :param df: The dataframe from which the matches were generated. Only needed if you want to store the ratings in the class object in which case the column names must also be passed.
        :param column_names: The column names of the dataframe. Only needed if you want to store the ratings in the class object in which case the df must also be passed.

        :return: A dataframe with the original columns + estimator_features_out, historical_features_out and non_estimator_rating_features_out
        """
        input_cols = df.columns.tolist()
        self.column_names = column_names
        if (
            self.column_names.participation_weight is not None
            and self.column_names.participation_weight not in df.columns
        ):
            raise ValueError(
                f"participation_weight {self.column_names.participation_weight} not in df columns"
            )

        matches = convert_df_to_matches(
            df=df,
            column_names=self.column_names,
            performance_column_name=self.performance_column,
            separate_player_by_position=self.seperate_player_by_position,
        )

        potential_feature_values = self._generate_potential_feature_values(
            matches=matches
        )

        df = df.assign(**potential_feature_values)
        known_features_out = known_features_out or self.known_features_return
        historical_features_out = (
            historical_features_out or self._historical_features_out
        )
        self._calculated_match_ids = df[self.column_names.match_id].unique().tolist()
        return df[list(set(input_cols + known_features_out + historical_features_out + self._non_estimator_rating_features_out))]

    def _generate_potential_feature_values(self, matches: list[Match]):
        pre_match_player_rating_values = []
        pre_match_team_rating_values = []
        pre_match_opponent_projected_rating_values = []
        pre_match_opponent_rating_values = []
        team_opponent_leagues = []
        rating_update_match_ids = []
        rating_update_team_ids = []
        rating_update_team_ids_opponent = []
        player_rating_changes = []
        player_leagues = []
        player_predicted_performances = []
        projected_participation_weights = []
        pre_match_team_projected_rating_values = []
        position_rating_difference_values = {}
        performances = []
        player_ids = []
        team_leagues = []

        team_rating_changes = []

        for match_idx, match in enumerate(matches):
            self._validate_match(match)

            pre_match_rating = PreMatchRating(
                id=match.id,
                teams=self._get_pre_match_team_ratings(match=match),
                day_number=match.day_number,
            )
            match_team_rating_changes = self._create_match_team_rating_changes(
                match=match, pre_match_rating=pre_match_rating
            )
            team_rating_changes += match_team_rating_changes

            if (
                match_idx == len(matches) - 1
                or matches[match_idx + 1].update_id != match.update_id
            ):
                self._update_ratings(team_rating_changes=team_rating_changes)
                team_rating_changes = []

            match_position_ratings = []

            for team_idx, team_rating_change in enumerate(match_team_rating_changes):
                match_position_ratings.append({})
                opponent_team = match_team_rating_changes[-team_idx + 1]
                for player_idx, player_rating_change in enumerate(
                    team_rating_change.players
                ):

                    position = match.teams[team_idx].players[player_idx].position
                    if position:
                        match_position_ratings[team_idx][
                            position
                        ] = player_rating_change.pre_match_rating_value

                    pre_match_team_projected_rating_values.append(
                        pre_match_rating.teams[team_idx].projected_rating_value
                    )
                    pre_match_opponent_projected_rating_values.append(
                        pre_match_rating.teams[-team_idx + 1].projected_rating_value
                    )

                    pre_match_player_rating_values.append(
                        player_rating_change.pre_match_rating_value
                    )
                    pre_match_team_rating_values.append(
                        pre_match_rating.teams[team_idx].rating_value
                    )
                    pre_match_opponent_rating_values.append(
                        pre_match_rating.teams[-team_idx + 1].rating_value
                    )
                    player_leagues.append(player_rating_change.league)
                    team_opponent_leagues.append(opponent_team.league)
                    team_leagues.append(team_rating_change.league)
                    rating_update_match_ids.append(match.update_id)
                    rating_update_team_ids.append(match.teams[team_idx].update_id)
                    rating_update_team_ids_opponent.append(
                        match.teams[-team_idx + 1].update_id
                    )
                    projected_participation_weights.append(
                        match.teams[team_idx]
                        .players[player_idx]
                        .performance.projected_participation_weight
                    )

                    performances.append(player_rating_change.performance)
                    player_predicted_performances.append(
                        player_rating_change.predicted_performance
                    )
                    player_rating_changes.append(
                        player_rating_change.rating_change_value
                    )
                    player_ids.append(player_rating_change.id)

            if self.distinct_positions:
                for team_idx in range(len(match_team_rating_changes)):
                    player_per_team_count = len(
                        match_team_rating_changes[team_idx].players
                    )

                    for position in self.distinct_positions:
                        if position not in position_rating_difference_values:
                            position_rating_difference_values[position] = []
                        if (
                            position in match_position_ratings[team_idx]
                            and position in match_position_ratings[-team_idx + 1]
                        ):
                            position_rating_difference_values[position] += [
                                match_position_ratings[team_idx][position]
                                - match_position_ratings[-team_idx + 1][position]
                            ] * player_per_team_count
                        else:
                            position_rating_difference_values[position] += [
                                0
                            ] * player_per_team_count

        potential_feature_values = self._get_shared_rating_values(
            position_rating_difference_values=position_rating_difference_values,
            pre_match_team_projected_rating_values=pre_match_team_projected_rating_values,
            pre_match_opponent_projected_rating_values=pre_match_opponent_projected_rating_values,
            pre_match_player_rating_values=pre_match_player_rating_values,
            player_leagues=player_leagues,
            team_opponent_leagues=team_opponent_leagues,
            projected_participation_weights=projected_participation_weights,
            match_ids=rating_update_match_ids,
            team_ids=rating_update_team_ids,
            team_id_opponents=rating_update_team_ids_opponent,
            player_ids=player_ids,
            team_leagues=team_leagues,
        )
        potential_feature_values[
            self.prefix + RatingHistoricalFeatures.PLAYER_RATING_DIFFERENCE
        ] = np.array(pre_match_player_rating_values) - np.array(
            pre_match_opponent_rating_values
        )
        potential_feature_values[
            self.prefix + RatingHistoricalFeatures.RATING_DIFFERENCE
        ] = np.array(pre_match_team_rating_values) - np.array(
            pre_match_opponent_rating_values
        )
        potential_feature_values[RatingKnownFeatures.PLAYER_RATING] = (
            pre_match_player_rating_values
        )
        potential_feature_values[
            self.prefix + RatingHistoricalFeatures.OPPONENT_RATING
        ] = pre_match_opponent_rating_values
        potential_feature_values[self.prefix + RatingHistoricalFeatures.TEAM_RATING] = (
            pre_match_team_rating_values
        )
        potential_feature_values[self.prefix + RatingHistoricalFeatures.RATING_MEAN] = (
            np.array(pre_match_team_rating_values) * 0.5
            + 0.5 * np.array(pre_match_opponent_rating_values)
        )

        potential_feature_values[
            self.prefix + RatingHistoricalFeatures.PLAYER_RATING_DIFFERENCE_FROM_TEAM
        ] = np.array(pre_match_player_rating_values) - np.array(
            pre_match_team_rating_values
        )
        potential_feature_values[self.prefix + RatingHistoricalFeatures.PERFORMANCE] = (
            performances
        )

        potential_feature_values[
            self.prefix + RatingHistoricalFeatures.PLAYER_RATING_CHANGE
        ] = player_rating_changes
        potential_feature_values[
            self.prefix + RatingHistoricalFeatures.PLAYER_PREDICTED_PERFORMANCE
        ] = player_predicted_performances

        return potential_feature_values

    def generate_future(
        self,
        df: Optional[pd.DataFrame],
        matches: Optional[list[Match]] = None,
        historical_features_out: Optional[list[RatingHistoricalFeatures]] = None,
        known_features_out: Optional[list[RatingKnownFeatures]] = None,
    ) -> pd.DataFrame:

        input_cols = df.columns.tolist()

        if (
            matches is not None
            and len(matches) > 0
            and not isinstance(matches[0], Match)
        ):
            raise ValueError("matches must be a list of Match objects")

        if matches is None and df is None:
            raise ValueError("If matches is not passed, df must be massed")

        if matches is None:
            matches = convert_df_to_matches(
                df=df,
                column_names=self.column_names,
                performance_column_name=self.performance_column,
                separate_player_by_position=self.seperate_player_by_position,
            )

        pre_match_player_rating_values = []
        pre_match_opponent_projected_rating_values = []
        team_opponent_leagues = []
        match_ids = []
        player_leagues = []
        projected_participation_weights = []
        rating_update_team_ids = []
        rating_update_team_ids_opponent = []
        player_ids = []
        team_leagues = []
        position_rating_difference_values = {}

        pre_match_team_projected_rating_values = []

        for match_idx, match in enumerate(matches):
            match_position_ratings = []
            self._validate_match(match)
            pre_match_rating = PreMatchRating(
                id=match.id,
                teams=self._get_pre_match_team_ratings(match=match),
                day_number=match.day_number,
            )

            for team_idx, pre_match_team in enumerate(pre_match_rating.teams):
                match_position_ratings.append({})
                opponent_team = pre_match_rating.teams[-team_idx + 1]
                for player_idx, pre_match_player in enumerate(pre_match_team.players):
                    position = match.teams[team_idx].players[player_idx].position
                    if position:
                        match_position_ratings[team_idx][
                            position
                        ] = pre_match_player.rating_value
                    pre_match_team_projected_rating_values.append(
                        pre_match_team.projected_rating_value
                    )
                    pre_match_player_rating_values.append(pre_match_player.rating_value)
                    pre_match_opponent_projected_rating_values.append(
                        opponent_team.projected_rating_value
                    )
                    team_opponent_leagues.append(opponent_team.league)
                    team_leagues.append(pre_match_team.league)
                    player_leagues.append(pre_match_player.league)
                    match_ids.append(match.id)
                    player_ids.append(pre_match_player.id)
                    rating_update_team_ids.append(match.teams[team_idx].update_id)
                    rating_update_team_ids_opponent.append(
                        match.teams[-team_idx + 1].update_id
                    )
                    projected_participation_weights.append(
                        match.teams[team_idx]
                        .players[player_idx]
                        .performance.projected_participation_weight
                    )

            if self.distinct_positions:
                for team_idx in range(len(pre_match_rating.teams)):
                    player_per_team_count = len(
                        pre_match_rating.teams[team_idx].players
                    )

                    for position in self.distinct_positions:
                        if position not in position_rating_difference_values:
                            position_rating_difference_values[position] = []
                        if (
                            position in match_position_ratings[team_idx]
                            and position in match_position_ratings[-team_idx + 1]
                        ):
                            position_rating_difference_values[position] += [
                                match_position_ratings[team_idx][position]
                                - match_position_ratings[-team_idx + 1][position]
                            ] * player_per_team_count
                        else:
                            position_rating_difference_values[position] += [
                                0
                            ] * player_per_team_count

        potential_feature_values = self._get_shared_rating_values(
            position_rating_difference_values=position_rating_difference_values,
            pre_match_team_projected_rating_values=pre_match_team_projected_rating_values,
            pre_match_opponent_projected_rating_values=pre_match_opponent_projected_rating_values,
            pre_match_player_rating_values=pre_match_player_rating_values,
            player_leagues=player_leagues,
            team_opponent_leagues=team_opponent_leagues,
            match_ids=match_ids,
            projected_participation_weights=projected_participation_weights,
            team_ids=rating_update_team_ids,
            team_id_opponents=rating_update_team_ids_opponent,
            player_ids=player_ids,
            team_leagues=team_leagues,
        )

        for f in self._historical_features_out:
            potential_feature_values[f] = [np.nan] * len(pre_match_player_rating_values)

        df = df.assign(**potential_feature_values)

        known_features_return = known_features_out or self.known_features_return
        historical_features_out = (
            historical_features_out or self._historical_features_out
        )
        return df[list(set(input_cols + known_features_return + historical_features_out + self._non_estimator_rating_features_out))]

    def _get_shared_rating_values(
        self,
        position_rating_difference_values: dict[str, list[float]],
        pre_match_team_projected_rating_values: list[float],
        pre_match_opponent_projected_rating_values: list[float],
        pre_match_player_rating_values: list[float],
        player_leagues: list[str],
        team_opponent_leagues: list[str],
        match_ids: list[str],
        team_ids: list[str],
        team_id_opponents: list[str],
        player_ids: list[str],
        projected_participation_weights: list[float],
        team_leagues: list[str],
    ) -> dict[Union[RatingKnownFeatures, RatingHistoricalFeatures], Any]:

        if self.column_names.projected_participation_weight:
            df = pd.DataFrame(
                {
                    "match_id": match_ids,
                    "team_id": team_ids,
                    "team_id_opponent": team_id_opponents,
                    RatingKnownFeatures.PLAYER_RATING: pre_match_player_rating_values,
                    "projected_participation_weight": projected_participation_weights,
                    "player_id": player_ids,
                }
            )

            game_player = (
                df.groupby(["match_id", "player_id", "team_id", "team_id_opponent"])[
                    [
                        RatingKnownFeatures.PLAYER_RATING,
                        "projected_participation_weight",
                    ]
                ]
                .mean()
                .reset_index()
            )

            game_player["game_team_sum_projected_participation_weight"] = (
                game_player.groupby(["match_id", "team_id"])[
                    "projected_participation_weight"
                ].transform("sum")
            )

            game_player["weighted_pre_match_player_rating_value"] = (
                game_player[self.prefix + RatingKnownFeatures.PLAYER_RATING]
                * game_player["projected_participation_weight"]
            )

            game_player[self.prefix + RatingKnownFeatures.TEAM_RATING_PROJECTED] = (
                game_player.groupby(["match_id", "team_id"])[
                    "weighted_pre_match_player_rating_value"
                ].transform("sum")
                / game_player["game_team_sum_projected_participation_weight"]
            )

            game_team = (
                game_player.groupby(["match_id", "team_id", "team_id_opponent"])[
                    self.prefix + RatingKnownFeatures.TEAM_RATING_PROJECTED
                ]
                .mean()
                .reset_index()
            )

            game_team = game_team.merge(
                game_team[
                    [
                        "match_id",
                        "team_id_opponent",
                        self.prefix + RatingKnownFeatures.TEAM_RATING_PROJECTED,
                    ]
                ].rename(
                    columns={
                        self.prefix
                        + RatingKnownFeatures.TEAM_RATING_PROJECTED: self.prefix
                        + RatingKnownFeatures.OPPONENT_RATING_PROJECTED
                    }
                ),
                left_on=["match_id", "team_id"],
                right_on=["match_id", "team_id_opponent"],
            )

            game_player = game_player.merge(
                game_team[
                    [
                        "match_id",
                        "team_id",
                        self.prefix + RatingKnownFeatures.OPPONENT_RATING_PROJECTED,
                    ]
                ],
                on=["match_id", "team_id"],
            )

            game_player[self.prefix + RatingKnownFeatures.RATING_MEAN_PROJECTED] = (
                game_player[self.prefix + RatingKnownFeatures.TEAM_RATING_PROJECTED]
                + game_player[
                    self.prefix + RatingKnownFeatures.OPPONENT_RATING_PROJECTED
                ]
            ) / 2

            df = df[["match_id", "player_id"]].merge(
                game_player[
                    [
                        "match_id",
                        "player_id",
                        self.prefix + RatingKnownFeatures.TEAM_RATING_PROJECTED,
                        self.prefix + RatingKnownFeatures.OPPONENT_RATING_PROJECTED,
                        self.prefix + RatingKnownFeatures.RATING_MEAN_PROJECTED,
                        self.prefix + RatingKnownFeatures.PLAYER_RATING,
                    ]
                ],
                on=["match_id", "player_id"],
                how="left",
            )

            rating_differences_projected = (
                df[self.prefix + RatingKnownFeatures.TEAM_RATING_PROJECTED]
                - df[self.prefix + RatingKnownFeatures.OPPONENT_RATING_PROJECTED]
            ).tolist()
            player_rating_difference_from_team_projected = (
                df[self.prefix + RatingKnownFeatures.PLAYER_RATING]
                - df[self.prefix + RatingKnownFeatures.TEAM_RATING_PROJECTED]
            ).tolist()
            player_rating_differences_projected = (
                df[self.prefix + RatingKnownFeatures.PLAYER_RATING]
                - df[self.prefix + RatingKnownFeatures.OPPONENT_RATING_PROJECTED]
            ).tolist()
            rating_means_projected = df[
                self.prefix + RatingKnownFeatures.RATING_MEAN_PROJECTED
            ].tolist()
            pre_match_opponent_projected_rating_values = df[
                self.prefix + RatingKnownFeatures.OPPONENT_RATING_PROJECTED
            ].tolist()
            pre_match_team_projected_rating_values = df[
                self.prefix + RatingKnownFeatures.TEAM_RATING_PROJECTED
            ].tolist()
            pre_match_player_rating_values = df[
                self.prefix + RatingKnownFeatures.PLAYER_RATING
            ].tolist()

        else:
            rating_differences_projected = (
                np.array(pre_match_team_projected_rating_values)
                - np.array(pre_match_opponent_projected_rating_values)
            ).tolist()
            player_rating_differences_projected = (
                np.array(pre_match_player_rating_values)
                - np.array(pre_match_opponent_projected_rating_values)
            ).tolist()
            player_rating_difference_from_team_projected = (
                np.array(pre_match_player_rating_values)
                - np.array(pre_match_team_projected_rating_values)
            ).tolist()
            rating_means_projected = (
                np.array(pre_match_team_projected_rating_values) * 0.5
                + 0.5 * np.array(pre_match_opponent_projected_rating_values)
            ).tolist()

        return_values = {
            self.prefix
            + RatingKnownFeatures.RATING_DIFFERENCE_PROJECTED: rating_differences_projected,
            self.prefix
            + RatingKnownFeatures.PLAYER_RATING_DIFFERENCE_FROM_TEAM_PROJECTED: player_rating_difference_from_team_projected,
            self.prefix
            + RatingKnownFeatures.PLAYER_RATING_DIFFERENCE_PROJECTED: player_rating_differences_projected,
            self.prefix
            + RatingKnownFeatures.TEAM_RATING_PROJECTED: pre_match_team_projected_rating_values,
            self.prefix
            + RatingKnownFeatures.OPPONENT_RATING_PROJECTED: pre_match_opponent_projected_rating_values,
            self.prefix
            + RatingKnownFeatures.PLAYER_RATING: pre_match_player_rating_values,
            self.prefix + RatingKnownFeatures.PLAYER_LEAGUE: player_leagues,
            self.prefix + RatingKnownFeatures.OPPONENT_LEAGUE: team_opponent_leagues,
            self.prefix + RatingKnownFeatures.TEAM_LEAGUE: team_leagues,
            self.prefix
            + RatingKnownFeatures.RATING_MEAN_PROJECTED: rating_means_projected,
            self.prefix + RatingKnownFeatures.MATCH_ID: match_ids,
        }

        if self.distinct_positions:
            for position, rating_values in position_rating_difference_values.items():
                return_values[
                    self.prefix
                    + RatingKnownFeatures.RATING_DIFFERENCE_POSITION
                    + "_"
                    + position
                ] = rating_values

        return return_values

    def _create_match_team_rating_changes(
        self, match: Match, pre_match_rating: PreMatchRating
    ) -> list[TeamRatingChange]:

        team_rating_changes = []

        for team_idx, pre_match_team_rating in enumerate(pre_match_rating.teams):
            team_rating_change = self.match_rating_generator.generate_rating_change(
                day_number=match.day_number,
                pre_match_team_rating=pre_match_team_rating,
                pre_match_opponent_team_rating=pre_match_rating.teams[-team_idx + 1],
            )
            team_rating_changes.append(team_rating_change)

        return team_rating_changes

    def _update_ratings(self, team_rating_changes: list[TeamRatingChange]):

        for idx, team_rating_change in enumerate(team_rating_changes):
            self.match_rating_generator.update_rating_by_team_rating_change(
                team_rating_change=team_rating_change,
                opponent_team_rating_change=team_rating_changes[-idx + 1],
            )

    def _get_pre_match_team_ratings(self, match: Match) -> list[PreMatchTeamRating]:
        pre_match_team_ratings = []
        for match_team in match.teams:
            pre_match_team_ratings.append(
                self.match_rating_generator.generate_pre_match_team_rating(
                    match_team=match_team, day_number=match.day_number
                )
            )

        return pre_match_team_ratings



from abc import abstractmethod, ABC
from narwhals.typing import FrameT

import pandas as pd
import narwhals as nw
import warnings

warnings.simplefilter(action="ignore", category=pd.errors.PerformanceWarning)


class PredictorTransformer(ABC):

    def __init__(self, features: list[str]):
        self.features = features

    @abstractmethod
    def fit_transform(self, df: FrameT) -> pd.DataFrame:
        pass

    @abstractmethod
    def transform(self, df: FrameT) -> pd.DataFrame:
        pass

    @property
    @abstractmethod
    def features_out(self) -> list[str]:
        pass


class ConvertDataFrameToCategoricalTransformer(PredictorTransformer):
    """
    Converts a specified list of columns to categorical dtype
    """

    def __init__(self, features: list[str]):
        super().__init__(features=features)

    @nw.narwhalify
    def fit_transform(self, df: FrameT) -> pd.DataFrame:
        self._features_out = self.features
        return self.transform(df.to_pandas())[self._features_out]

    @nw.narwhalify
    def transform(self, df: FrameT) -> pd.DataFrame:
        df = df.with_columns(
            nw.col(feature).cast(nw.Categorical) for feature in self.features
        )
        return df.select(self._features_out).to_pandas()

    @property
    def features_out(self) -> list[str]:
        return self._features_out


class SkLearnTransformerWrapper(PredictorTransformer):
    """
    A wrapper around an Sklearn Transformer
    """

    def __init__(self, transformer, features: list[str]):
        self.transformer = transformer
        super().__init__(features=features)
        self._features_out = []

    @nw.narwhalify
    def fit_transform(self, df: FrameT) -> pd.DataFrame:

        try:
            transformed_values = self.transformer.fit_transform(
                df.select(self.features).to_pandas()
            ).toarray()
        except AttributeError:
            transformed_values = self.transformer.fit_transform(df.select(self.features))
            if isinstance(transformed_values, pd.DataFrame):
                transformed_values = transformed_values.to_numpy()

        self._features_out = self.transformer.get_feature_names_out().tolist()


        return df.to_pandas().assign(
            **{
                self._features_out[idx]: transformed_values[:, idx]
                for idx in range(len(self._features_out))
            }
        )[self._features_out]



    def transform(self, df: FrameT) -> pd.DataFrame:
        try:
            transformed_values = self.transformer.transform(df.select(self.features)).toarray()
        except AttributeError:
            transformed_values = self.transformer.transform(df.select(self.features))
            if isinstance(transformed_values, pd.DataFrame):
                transformed_values = transformed_values.to_numpy()


        return df.to_pandas().assign(
            **{
                self._features_out[idx]: transformed_values[:, idx]
                for idx in range(len(self._features_out))
            }
        )[self._features_out]

    @property
    def features_out(self) -> list[str]:
        return self._features_out


import pandas as pd

from player_performance_ratings.pipeline import Pipeline
from player_performance_ratings.predictor import GameTeamPredictor
from player_performance_ratings.tuner.performances_generator_tuner import (
    PerformancesSearchRange,
)
from player_performance_ratings.tuner.predictor_tuner import PredictorTuner

from player_performance_ratings.tuner.rating_generator_tuner import (
    UpdateRatingGeneratorTuner,
)
from player_performance_ratings.ratings import UpdateRatingGenerator

from player_performance_ratings.data_structures import ColumnNames

from player_performance_ratings.tuner import PipelineTuner, PerformancesGeneratorTuner
from player_performance_ratings.tuner.utils import (
    ParameterSearchRange,
    get_default_team_rating_search_range,
)

column_names = ColumnNames(
    team_id="teamname",
    match_id="gameid",
    start_date="date",
    player_id="playername",
    league="league",
    position="position",
)
df = pd.read_parquet("data/subsample_lol_data")
df = df.sort_values(by=["date", "gameid", "teamname", "playername"])
df["champion_position"] = df["champion"] + df["position"]
df["__target"] = df["result"]

df = df.drop_duplicates(subset=["gameid", "teamname", "playername"])

df = (
    df.assign(team_count=df.groupby("gameid")["teamname"].transform("nunique"))
    .loc[lambda x: x.team_count == 2]
    .drop(columns=["team_count"])
)
df = df.drop_duplicates(subset=["gameid", "teamname", "playername"])

rating_generator = UpdateRatingGenerator(performance_column="performance")

predictor = GameTeamPredictor(
    game_id_colum="gameid",
    team_id_column="teamname",
)

pipeline = Pipeline(
    rating_generators=rating_generator, predictor=predictor, column_names=column_names
)

start_rating_search_range = [
    ParameterSearchRange(
        name="team_weight",
        type="uniform",
        low=0.12,
        high=0.4,
    ),
    ParameterSearchRange(
        name="league_quantile",
        type="uniform",
        low=0.12,
        high=0.4,
    ),
    ParameterSearchRange(
        name="min_count_for_percentiles",
        type="uniform",
        low=20,
        high=100,
    ),
]

performance_generator_tuner = PerformancesGeneratorTuner(
    performances_search_range=PerformancesSearchRange(
        search_ranges=[
            ParameterSearchRange(
                name="damagetochampions", type="uniform", low=0, high=0.45
            ),
            ParameterSearchRange(
                name="deaths", type="uniform", low=0, high=0.3, lower_is_better=True
            ),
            ParameterSearchRange(name="kills", type="uniform", low=0, high=0.3),
            ParameterSearchRange(name="result", type="uniform", low=0.25, high=0.85),
        ]
    ),
    n_trials=3,
)

rating_generator_tuner = UpdateRatingGeneratorTuner(
    team_rating_search_ranges=get_default_team_rating_search_range(),
    start_rating_search_ranges=start_rating_search_range,
    optimize_league_ratings=True,
    team_rating_n_trials=3,
)

tuner = PipelineTuner(
    performances_generator_tuners=performance_generator_tuner,
    rating_generator_tuners=rating_generator_tuner,
    predictor_tuner=PredictorTuner(
        n_trials=1,
        search_ranges=[
            ParameterSearchRange(name="C", type="categorical", choices=[1.0, 0.5])
        ],
    ),
    fit_best=True,
    pipeline=pipeline,
)
best_match_predictor, df = tuner.tune(
    df=df, return_df=True, return_cross_validated_predictions=True
)



from abc import abstractmethod, ABC
from typing import Literal, Optional

import pandas as pd
from narwhals.typing import FrameT
import narwhals as nw


class SimpleTransformer(ABC):

    @abstractmethod
    def transform(self, df: FrameT) -> pd.DataFrame:
        pass


class AggregatorTransformer(SimpleTransformer):

    def __init__(self, columns: list[str],column_to_alias: Optional[dict[str, str]] = None, granularity: Optional[list] = None,
                 aggregator: Literal['sum', 'mean'] = 'sum'):
        self.columns = columns
        self.column_to_alias = column_to_alias or {f: f"{f}_{aggregator}" for f in columns}
        self.granularity = granularity
        self.aggregator = aggregator

    @nw.narwhalify
    def transform(self, df: FrameT) -> pd.DataFrame:
        if self.aggregator == 'sum':
            if self.granularity:
                return df.with_columns(
                    nw.col(column)
                    .sum()
                    .over(self.granularity)
                    .alias(self.column_to_alias[column])
                    for column in self.columns
                ).to_pandas()
            else:
                return df.with_columns(
                    nw.col(column)
                    .sum()
                    .alias(self.column_to_alias[column])
                    for column in self.columns
                ).to_pandas()

        elif self.aggregator == 'mean':
            if self.granularity:
                return df.with_columns(
                    nw.col(column)
                    .mean()
                    .over(self.granularity)
                    .alias(self.column_to_alias[column])
                    for column in self.columns
                ).to_pandas()
            else:
                return df.with_columns(
                    nw.col(column)
                    .mean()
                    .alias(self.column_to_alias[column])
                    for column in self.columns
                ).to_pandas()

        raise NotImplementedError(f"Aggregator {self.aggregator} not implemented")


class NormalizerToColumnTransformer(SimpleTransformer):

    def __init__(self, column: str, granularity: list[str], normalize_to_column: str):
        self.column = column
        self.granularity = granularity
        self.normalize_to_column = normalize_to_column

    @nw.narwhalify
    def transform(self, df: FrameT) -> pd.DataFrame:
        input_cols = df.columns
        df = df.with_columns(
            nw.col(self.column)
            .sum()
            .over(self.granularity)
            .alias('__sum_value')
        )
        return df.with_columns(
            (nw.col(self.column) / nw.col('__sum_value') * nw.col(self.normalize_to_column)).alias(self.column)
        ).select(input_cols).to_pandas()



from .base_tuner import BaseTuner
from .performances_generator_tuner import PerformancesGeneratorTuner
from .pipeline_tuner import PipelineTuner
from .rating_generator_tuner import UpdateRatingGeneratorTuner
from .start_rating_optimizer import StartLeagueRatingOptimizer



import pandas as pd

from player_performance_ratings import ColumnNames
import polars as pl


def convert_pandas_to_polars(df: pd.DataFrame) -> pl.DataFrame:
    for column in df.select_dtypes(include=['category']).columns:
        df[column] = df[column].astype(str)

    return pl.from_pandas(df)


def validate_sorting(df: pd.DataFrame, column_names: ColumnNames) -> None:
    df_sorted = df.sort_values(
        by=[
            column_names.start_date,
            column_names.match_id,
            column_names.team_id,
            column_names.player_id,
        ]
    )

    if not df.equals(df_sorted):
        for column in [
            column_names.match_id,
            column_names.team_id,
            column_names.player_id,
        ]:
            df = df.assign(**{column: df[column].astype("str")})

        df_sorted = df.sort_values(
            by=[
                column_names.start_date,
                column_names.match_id,
                column_names.team_id,
                column_names.player_id,
            ]
        )

        if df.equals(df_sorted):
            return
        for column in [
            column_names.match_id,
            column_names.team_id,
            column_names.player_id,
        ]:
            try:
                df = df.assign(**{column: df[column].astype("int")})
            except:
                pass

        df_sorted = df.sort_values(
            by=[
                column_names.start_date,
                column_names.match_id,
                column_names.team_id,
                column_names.player_id,
            ]
        )

        if df.equals(df_sorted):
            return

        raise ValueError(
            "df needs to be sorted by date, game_id, team_id, player_id in ascending order"
        )



from typing import Optional, Union, List

import pandas as pd
import polars as pl

from player_performance_ratings import ColumnNames
from player_performance_ratings.pipeline import DataFrameType
from player_performance_ratings.ratings import convert_df_to_matches, LeagueIdentifier
from player_performance_ratings.ratings.performance_generator import (
    PerformancesGenerator,
)
from player_performance_ratings.ratings.rating_generator import RatingGenerator
from player_performance_ratings.transformers.base_transformer import (
    BaseLagGenerator,
    BaseLagGeneratorPolars,
    BaseTransformer,
)


class PipelineTransformer:
    """
    Pipeline of rating_generators, lag_generators and transformers to be applied to a dataframe
    For historical data use fit_transform
    For future data use transform.
    """

    def __init__(
        self,
        column_names: ColumnNames,
        performances_generator: Optional[PerformancesGenerator] = None,
        rating_generators: Optional[
            Union[RatingGenerator, list[RatingGenerator]]
        ] = None,
        pre_lag_transformers: Optional[list[BaseTransformer]] = None,
        lag_generators: Optional[
            List[Union[BaseLagGenerator, BaseLagGeneratorPolars]]
        ] = None,
        post_lag_transformers: Optional[list[BaseTransformer]] = None,
    ):
        self.column_names = column_names
        self.performances_generator = performances_generator
        self.rating_generators = rating_generators
        if rating_generators is None:
            self.rating_generators = []
        if isinstance(rating_generators, RatingGenerator):
            self.rating_generators = [rating_generators]

        self.pre_lag_transformers = pre_lag_transformers or []
        self.lag_generators = lag_generators or []
        self.post_lag_transformers = post_lag_transformers or []

    def fit_transform(self, df: DataFrameType) -> DataFrameType:
        """
        Fit and transform the pipeline on historical data
        :param df: Either polars or Pandas dataframe
        """

        original_is_polars = False
        if isinstance(df, pl.DataFrame):
            original_is_polars = True
            df = df.to_pandas()

        if self.performances_generator:
            df = self.performances_generator.generate(df)

        if self.rating_generators:
            matches = convert_df_to_matches(
                column_names=self.column_names,
                df=df,
                league_identifier=LeagueIdentifier(),
                performance_column_name=self.rating_generators[0].performance_column,
            )
        else:
            matches = []
        for rating_idx, rating_generator in enumerate(self.rating_generators):
            match_ratings = rating_generator.generate_historical_by_matches(
                column_names=self.column_names, matches=matches
            )
            for rating_feature, values in match_ratings.items():
                if len(self.rating_generators) > 1:
                    rating_feature_str = rating_feature + str(rating_idx)
                else:
                    rating_feature_str = rating_feature

                df = df.assign(**{rating_feature_str: values})

        for transformer in self.pre_lag_transformers:
            df = transformer.fit_transform(df=df, column_names=self.column_names)

        for lag_generator in self.lag_generators:
            df = lag_generator.generate_historical(
                df=df, column_names=self.column_names
            )
        if original_is_polars:
            df = pl.DataFrame(df)
        return df

    def transform(self, df: DataFrameType) -> DataFrameType:
        """
        Transform the pipeline on future data
        :param df: Either polars or Pandas dataframe
        """

        original_is_polars = False
        if isinstance(df, pl.DataFrame):
            original_is_polars = True
            df = df.to_pandas()

        if self.performances_generator:
            df = self.performances_generator.generate(df)

        if self.rating_generators:
            if self.rating_generators[0].performance_column not in df.columns:
                raise ValueError(f"Performance column {self.rating_generators[0].performance_column} not found in dataframe")
            matches = convert_df_to_matches(
                column_names=self.column_names,
                df=df,
                league_identifier=LeagueIdentifier(),
                performance_column_name=self.rating_generators[0].performance_column,
            )
        else:
            matches = []

        for rating_idx, rating_generator in enumerate(self.rating_generators):
            df = rating_generator.generate_future(df=df, matches=matches)

        for transformer in self.pre_lag_transformers:
            df = transformer.transform(df)

        for lag_generator in self.lag_generators:
            df = lag_generator.generate_future(df)

        for transformer in self.post_lag_transformers:
            df = transformer.transform(df)

        if original_is_polars:
            df = pl.DataFrame(df)

        return df



from setuptools import setup, find_packages

with open("requirements.txt") as f:
    required = f.read().splitlines()

setup(
    name="player-performance-ratings",
    version="5.16",
    packages=find_packages(),
    install_requires=required,
    author="Mathias Holmstrøm",
    author_email="mathiasholmstom@gmail.com",
    description="Match Predictions based on Player Ratings",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/Hiderdk/player-performance-ratings",
)



import logging
from typing import List, Optional, Union, TypeVar

import pandas as pd
import polars as pl
import narwhals as nw
from sklearn.metrics import log_loss, mean_absolute_error

from player_performance_ratings.scorer import SklearnScorer, OrdinalLossScorer

from player_performance_ratings.cross_validator.cross_validator import (
    CrossValidator,
    MatchKFoldCrossValidator,
)
from player_performance_ratings.ratings.performance_generator import (
    PerformancesGenerator,
)

from player_performance_ratings.consts import PredictColumnNames
from player_performance_ratings.predictor._base import BasePredictor

from player_performance_ratings.data_structures import Match, ColumnNames
from player_performance_ratings.ratings.league_identifier import LeagueIdentifier
from player_performance_ratings.ratings.match_generator import convert_df_to_matches
from player_performance_ratings.ratings.rating_generator import RatingGenerator

from player_performance_ratings.transformers.base_transformer import (
    BaseTransformer,
    BaseLagGenerator,
    BaseLagGeneratorPolars,
)
from player_performance_ratings.utils import convert_pandas_to_polars

DataFrameType = TypeVar("DataFrameType", pd.DataFrame, pl.DataFrame)


class Pipeline:
    """
    Pipeline class for generating predictions on a dataset using a rating generators, lag generators, and transformers that feeds into a Predictor.
    The pipeline ensures the training process and prediction process is consistent across the entire end-to-end feature engineering and prediction process.
    Another advantage of using the pipeline over the transformers, lag_generators and predictor separately is that the estimator_features are automatically updated
    The output of the transformers, lag_generators are automatically added to the estimator_features of the predictor - the user does not need to add these as estimator_features.
    The only estimator_features required by the user to manually pass into the predictor are the features that are not generated by the pipeline.

    Use .train_predict() to train the pipeline on a dataset and generate predictions.
    Use .future_predict() to generate predictions on a future dataset.


    Further, cross-validation is supported by the pipeline.
        Thus, if the output of the predictions should be cross-validated that can be done using the cross_validate_predict() method.
        Alternatively, cross_validate can be set to True when calling .train_predict().
         This will both train a pipeline on all historical data and return cross-validated predictions.

    """

    def __init__(
        self,
        predictor: BasePredictor,
        column_names: ColumnNames,
        performances_generator: Optional[PerformancesGenerator] = None,
        rating_generators: Optional[
            Union[RatingGenerator, list[RatingGenerator]]
        ] = None,
        pre_lag_transformers: Optional[list[BaseTransformer]] = None,
        lag_generators: Optional[
            List[Union[BaseLagGenerator, BaseLagGeneratorPolars]]
        ] = None,
        post_lag_transformers: Optional[list[BaseTransformer]] = None,
    ):
        """
        :param predictor: The predictor to use for generating the predictions
        :param column_names:
        :param performances_generator:
            An optional transformer class that take place in order to convert one or multiple column names into the performance value that is used by the rating model
        :param rating_generators:      A single or a list of RatingGenerators.
        :param pre_lag_transformers:   A list of transformers that take place before the lag generators
        :param lag_generators:        A list of lag generators that generate lags, rolling-means
        :param post_lag_transformers: A list of transformers that take place after the lag generators.
            This makes it possble to transform the lagged features before they are used by the predictor.
        """

        self._estimator_features = predictor._estimator_features
        self.rating_generators: list[RatingGenerator] = (
            rating_generators
            if isinstance(rating_generators, list)
            else [rating_generators]
        )
        if rating_generators is None:
            self.rating_generators: list[RatingGenerator] = []

        self.pre_lag_transformers = pre_lag_transformers or []
        self.post_lag_transformers = post_lag_transformers or []
        self.lag_generators = lag_generators or []
        self.column_names = column_names

        est_feats = predictor.estimator_features
        for r in self.rating_generators:
            est_feats = list(set(est_feats + r.known_features_return))
        for f in self.lag_generators:
            est_feats = list(set(est_feats +f.estimator_features_out))
        for idx, post_transformer in enumerate(self.post_lag_transformers):
            if hasattr(post_transformer, "predictor") and not post_transformer.features:
                self.post_lag_transformers[idx].features = est_feats.copy()
            est_feats = list(set(est_feats +  self.post_lag_transformers[idx].estimator_features_out))

        for c in [
            *self.lag_generators,
            *self.pre_lag_transformers,
            *self.post_lag_transformers,
        ]:
            self._estimator_features += [
                f for f in c.estimator_features_out if f not in self._estimator_features
            ]
        for rating_idx, c in enumerate(self.rating_generators):
            for rating_feature in c.known_features_out:
                if rating_feature not in self._estimator_features:
                    self._estimator_features.append(rating_feature)

        logging.info(f"Using estimator features {self._estimator_features}")
        self.performances_generator = performances_generator
        self.predictor = predictor
        self.predictor.set_target(PredictColumnNames.TARGET)

    def cross_validate_score(
        self,
        df: DataFrameType,
        cross_validator: Optional[CrossValidator] = None,
        matches: Optional[list[Match]] = None,
        create_performance: bool = True,
        create_rating_features: bool = True,
    ) -> float:
        """
        Calculates the cross-validation score for the pipeline.
        :param df: DataFrame with the data to be used for cross-validation
        :param cross_validator: CrossValidator object to be used for cross-validation
        :param matches: If list of matches are provided, these will be used for rating generation.
        If not provided, the matches will be generated from the df if rating-generation take place during the pipeline.
        :param create_performance: If True, the performance generator will be used to generate performance values and add it to the dataframe.
        :param create_rating_features: If True, the rating generator will be used to generate rating values and add it to the dataframe.
        """

        for col in self.predictor.columns_added:
            if col in df.columns:
                df = df.drop(columns=[col])

        if cross_validator is None:
            cross_validator = self._create_default_cross_validator(df=df)

        if create_performance:
            df = self._add_performance(df=df)

        for rating_generator in self.rating_generators:
            create_rating_features = any(
                feature not in df.columns
                for feature in rating_generator.known_features_return
            )
            if create_rating_features:
                break

        if create_rating_features and self.rating_generators:
            df = self._add_rating(matches=matches, df=df)

        validation_df = cross_validator.generate_validation_df(
            df=df,
            predictor=self.predictor,
            column_names=self.column_names,
            post_lag_transformers=self.post_lag_transformers,
            pre_lag_transformers=self.pre_lag_transformers,
            lag_generators=self.lag_generators,
            estimator_features=self._estimator_features,
            return_features=False,
        )

        if cross_validator.scorer is None:
            scorer = self._create_default_scorer(df)
            return cross_validator.cross_validation_score(
                validation_df=validation_df, scorer=scorer
            )

        return cross_validator.cross_validation_score(validation_df=validation_df)

    def cross_validate_predict(
        self,
        df: DataFrameType,
        cross_validator: Optional[CrossValidator] = None,
        matches: Optional[list[Match]] = None,
        create_performance: bool = True,
        create_rating_features: bool = True,
        return_features: bool = False,
        add_train_prediction: bool = False,
    ) -> DataFrameType:
        """
        Generates predictions on the validation dataset from the entire pipeline

        :param df: DataFrame with the data to be used for cross-validation
        :param cross_validator: CrossValidator object to be used for cross-validation
            If not set, a default MatchKFoldCrossValidator will be used
        :param matches: If list of matches are provided, these will be used for rating generation.
            If not provided, the matches will be generated from the df if rating-generation take place during the pipeline.
        :param create_performance: If True, the performance generator will be used to generate performance values and add it to the dataframe.
        :param create_rating_features: If True, the rating generator will be used to generate rating values and add it to the dataframe.
        :param return_features: If True, the features generated by the pipeline will be returned in the output dataframe.
        :param add_train_prediction: If True, the predictions on the training dataset will be added to the output dataframe.
        """
        if isinstance(df, pl.DataFrame) or isinstance(df, nw.DataFrame):
            df = df.to_pandas()
        cross_validated_df = df.copy()
        if cross_validator is None:
            cross_validator = self._create_default_cross_validator(
                df=cross_validated_df
            )

        if self.predictor.target not in cross_validated_df.columns:
            raise ValueError(
                f"Target {self.predictor.target} not in df columns. Target always needs to be set equal to {PredictColumnNames.TARGET}"
            )

        if create_performance:
            cross_validated_df = self._add_performance(df=cross_validated_df)

        for rating_generator in self.rating_generators:
            create_rating_features = any(
                feature not in df.columns
                for feature in rating_generator.known_features_return
            )
            if create_rating_features:
                break

        if create_rating_features and self.rating_generators:
            if self.rating_generators[0].performance_column not in cross_validated_df.columns.tolist():
                raise ValueError(
                    f"Performance column {self.rating_generators[0].performance_column} not found in dataframe")
            cross_validated_df = self._add_rating(
                matches=matches, df=cross_validated_df
            )

        cross_validated_df = cross_validator.generate_validation_df(
            df=cross_validated_df,
            predictor=self.predictor,
            column_names=self.column_names,
            lag_generators=self.lag_generators,
            post_lag_transformers=self.post_lag_transformers,
            pre_lag_transformers=self.pre_lag_transformers,
            estimator_features=self._estimator_features,
            return_features=return_features,
            add_train_prediction=add_train_prediction,
        )

        cn = self.column_names
        for _, row in (
            df[[cn.match_id, cn.team_id, cn.player_id]].dtypes.reset_index().iterrows()
        ):
            cross_validated_df[row["index"]] = cross_validated_df[row["index"]].astype(
                row[0]
            )
        if return_features:
            cols_to_drop = []
            for c in list(set(self._estimator_features + self.predictor.columns_added)):
                if c in cross_validated_df.columns and c in df.columns:
                    cols_to_drop.append(c)
            df = df.drop(columns=cols_to_drop)
            new_feats = [f for f in cross_validated_df.columns if f not in df.columns]
            return df.merge(
                cross_validated_df[new_feats + [cn.match_id, cn.team_id, cn.player_id]],
                on=[cn.match_id, cn.team_id, cn.player_id],
                how="left",
            )

        predictor_cols_added = self.predictor.columns_added
        if (
            "classes" in cross_validated_df.columns
            and "classes" not in predictor_cols_added
            and "classes" not in df.columns
        ):
            predictor_cols_added.append("classes")

        return df.merge(
            cross_validated_df[
                predictor_cols_added + [cn.match_id, cn.team_id, cn.player_id, cross_validator.validation_column_name]
            ],
            on=[cn.match_id, cn.team_id, cn.player_id],
            how="left",
        )

    def _create_default_cross_validator(self, df: pd.DataFrame) -> CrossValidator:

        scorer = self._create_default_scorer(df)

        return MatchKFoldCrossValidator(
            date_column_name=self.column_names.start_date,
            match_id_column_name=self.column_names.update_match_id,
            scorer=scorer,
        )

    def _create_default_scorer(self, df: pd.DataFrame):
        if self.predictor.estimator_type == "regressor":
            scorer = SklearnScorer(
                scorer_function=mean_absolute_error,
                pred_column=self.predictor.pred_column,
            )
            logging.info("Using mean_absolute_error as scorer")
        else:
            if len(df[PredictColumnNames.TARGET].unique()) > 2:
                scorer = OrdinalLossScorer(pred_column=self.predictor.pred_column)
                logging.info("Using ordinal loss as scorer")
            else:
                scorer = SklearnScorer(
                    scorer_function=log_loss, pred_column=self.predictor.pred_column
                )
                logging.info("Using log_loss as scorer")

        return scorer

    def train_predict(
        self,
        df: Union[pd.DataFrame, pl.DataFrame],
        matches: Optional[Union[list[Match], list[list[Match]]]] = None,
        return_features: bool = False,
        cross_validate_predict: bool = False,
        cross_validator: Optional[CrossValidator] = None,
    ) -> DataFrameType:
        """
        Trains the pipeline on the given dataframe and generates and returns predictions.

        :param df: DataFrame with the data to be used for training and prediction
        :param matches: If list of matches are provided, these will be used for rating generation.
            If not provided, the matches will be generated from the df if rating-generation take place during the pipeline.
        :param return_features: If True, the features generated by the pipeline will be returned in the output dataframe.
        :param cross_validate_predict: If True, the predictions will be generated using cross-validation.
        :param cross_validator: CrossValidator object to be used for cross-validation.
            If not set and cross_validate_predict is True, a default MatchKFoldCrossValidator will be used.
            Will have no impact if cross_validate_predict is False.

        """
        self.reset_pipeline()
        if isinstance(df, pl.DataFrame):
            df = df.to_pandas()
        df_with_predict = df.copy()

        if self.predictor.target not in df_with_predict.columns:
            raise ValueError(
                f"Target {self.predictor.target} not in df columns. Target always needs to be set equal to {PredictColumnNames.TARGET}"
            )

        ori_cols = df.columns.tolist()
        df_with_predict = self._add_performance(df=df_with_predict)
        if self.rating_generators:
            if self.rating_generators[0].performance_column not in df_with_predict.columns.tolist():
                raise ValueError(
                    f"Performance column {self.rating_generators[0].performance_column} not found in dataframe")
            df_with_predict = self._add_rating(
                matches=matches,
                df=df_with_predict,
            )

        if cross_validate_predict:
            cols = df_with_predict.columns.tolist()
            df_cv_predict = self.cross_validate_predict(
                df=df_with_predict,
                return_features=return_features,
                create_rating_features=False,
                create_performance=False,
                add_train_prediction=True,
                cross_validator=cross_validator,
            )
            cv_cols_added = [c for c in df_cv_predict.columns if c not in cols]
        else:
            cv_cols_added = []

        for idx in range(len(self.pre_lag_transformers)):
            self.pre_lag_transformers[idx].reset()
            df_with_predict = self.pre_lag_transformers[idx].fit_transform(
                df_with_predict, column_names=self.column_names
            )

        for idx in range(len(self.lag_generators)):
            self.lag_generators[idx].reset()

            count_remaining_polars = [
                l for l in self.lag_generators[idx:] if "Polars" in l.__class__.__name__
            ] + [
                l
                for l in self.post_lag_transformers
                if "Polars" in l.__class__.__name__
            ]

            if isinstance(df_with_predict, pd.DataFrame) and len(
                count_remaining_polars
            ) == len(self.lag_generators[idx:] + self.post_lag_transformers):
                df_with_predict = convert_pandas_to_polars(df_with_predict)

            df_with_predict = self.lag_generators[idx].generate_historical(
                df_with_predict, column_names=self.column_names
            )
        for idx in range(len(self.post_lag_transformers)):
            self.post_lag_transformers[idx].reset()
            df_with_predict = self.post_lag_transformers[idx].fit_transform(
                df_with_predict, column_names=self.column_names
            )

        if isinstance(df_with_predict, pl.DataFrame):
            df_with_predict = df_with_predict.to_pandas()

        self.predictor.train(
            df=df_with_predict, estimator_features=self._estimator_features
        )

        if cross_validate_predict:
            df_with_predict = df_cv_predict
        else:
            df_with_predict = self.predictor.add_prediction(df=df_with_predict)
        cn = self.column_names
        for _, row in (
            df[[cn.match_id, cn.team_id, cn.player_id]].dtypes.reset_index().iterrows()
        ):
            df_with_predict[row["index"]] = df_with_predict[row["index"]].astype(row[0])

        if return_features:
            new_feats = [f for f in df_with_predict.columns if f not in ori_cols]
            return df.merge(
                df_with_predict[new_feats + [cn.match_id, cn.team_id, cn.player_id]],
                on=[cn.match_id, cn.team_id, cn.player_id],
                how="left",
            )

        predictor_cols_added = self.predictor.columns_added
        if (
            "classes" in df_with_predict.columns
            and "classes" not in predictor_cols_added
            and "classes" not in df.columns
        ):
            predictor_cols_added.append("classes")

        return df.merge(
            df_with_predict[
                predictor_cols_added + [cn.match_id, cn.team_id, cn.player_id] + [c for c in cv_cols_added if c not in predictor_cols_added]
            ],
            on=[cn.match_id, cn.team_id, cn.player_id],
            how="left",
        )

    def reset_pipeline(self):
        for idx in range(len(self.rating_generators)):
            self.rating_generators[idx].reset_ratings()

        for transformer in [
            *self.lag_generators,
            *self.pre_lag_transformers,
            *self.post_lag_transformers,
        ]:
            transformer.reset()

    def _add_performance(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df.copy()

        if self.predictor.pred_column in df.columns:
            raise ValueError(
                f"Predictor column {self.predictor.pred_column} already in df columns. Remove or rename before generating predictions"
            )

        elif self.performances_generator:
            df = self.performances_generator.generate(df)

        if self.predictor.target not in df.columns:
            raise ValueError(
                f"Target {self.predictor.target} not in df columns. Target always needs to be set equal to {PredictColumnNames.TARGET}"
            )

        return df

    def _add_rating(
        self,
        matches: Optional[Union[list[Match], Match]],
        df: pd.DataFrame,
    ):

        if matches:
            if isinstance(matches[0], Match):
                matches = [matches for _ in self.rating_generators]

        rg = self.rating_generators[0]
        match_ids_calculated = rg.calculated_match_ids
        not_calculated_match_ids = (
            df[~df[self.column_names.match_id].isin(match_ids_calculated)][
                self.column_names.match_id
            ]
            .unique()
            .tolist()
        )

        df_no_ratings = df[~df[self.column_names.match_id].isin(match_ids_calculated)]
        for rating_idx, rating_generator in enumerate(self.rating_generators):
            if len(df_no_ratings) > 0:

                if matches is None:
                    rating_matches = convert_df_to_matches(
                        column_names=self.column_names,
                        df=df_no_ratings,
                        league_identifier=LeagueIdentifier(),
                        performance_column_name=rating_generator.performance_column,
                    )
                else:
                    rating_matches = matches[rating_idx]
                    if len(df_no_ratings) != len(df):
                        rating_matches = [
                            m
                            for m in rating_matches
                            if m.id in not_calculated_match_ids
                        ]

                match_ratings = rating_generator.generate_historical_by_matches(
                    matches=rating_matches, column_names=self.column_names
                )

                for rating_feature, values in match_ratings.items():
                    rating_feature_str = rating_feature
                    df_no_ratings[rating_feature_str] = values

        df = df_no_ratings

        return df

    def future_predict(
            self, df: DataFrameType, return_features: bool = False, return_rating_features: bool = False
    ) -> DataFrameType:
        """
        Generates predictions on a future dataset from the entire pipeline

        :param df: DataFrame with the data to be used for training and prediction
        :param return_features: If True, the features generated by the pipeline will be returned in the output dataframe.
        """
        if isinstance(df, pl.DataFrame):
            df = df.to_pandas()
        df_with_predict = df.copy()

        for rating_idx, rating_generator in enumerate(self.rating_generators):
            if rating_generator.performance_column in df_with_predict.columns:
                df_with_predict = df_with_predict.drop(
                    columns=[rating_generator.performance_column]
                )
            rating_column_names = rating_generator.column_names

            matches = convert_df_to_matches(
                column_names=rating_column_names,
                df=df_with_predict,
                league_identifier=LeagueIdentifier(),
                performance_column_name=rating_generator.performance_column,
            )

            df_with_predict = rating_generator.generate_future(
                matches=matches, df=df_with_predict
            )

        for pre_lag_transformer in self.pre_lag_transformers:
            df_with_predict = pre_lag_transformer.transform(df_with_predict)
        for idx, lag_generator in enumerate(self.lag_generators):
            count_remaining_polars = [
                                         l for l in self.lag_generators[idx:] if "Polars" in l.__class__.__name__
                                     ] + [
                                         l
                                         for l in self.post_lag_transformers
                                         if "Polars" in l.__class__.__name__
                                     ]

            if isinstance(df_with_predict, pd.DataFrame) and len(
                    count_remaining_polars
            ) == len(self.lag_generators[idx:] + self.post_lag_transformers):
                df_with_predict = convert_pandas_to_polars(df_with_predict)
            df_with_predict = lag_generator.generate_future(df_with_predict)
        for post_lag_transformer in self.post_lag_transformers:
            df_with_predict = post_lag_transformer.transform(df_with_predict)
        if isinstance(df_with_predict, pl.DataFrame):
            df_with_predict = df_with_predict.to_pandas()
        df_with_predict = self.predictor.add_prediction(df_with_predict)

        cn = self.column_names
        for _, row in (
                df[[cn.match_id, cn.team_id, cn.player_id]].dtypes.reset_index().iterrows()
        ):
            df_with_predict[row["index"]] = df_with_predict[row["index"]].astype(row[0])
        if return_features:
            new_feats = [f for f in df_with_predict.columns if f not in df.columns]
            return df.merge(
                df_with_predict[new_feats + [cn.match_id, cn.team_id, cn.player_id]],
                on=[cn.match_id, cn.team_id, cn.player_id],
                how="left",
            )
        elif return_rating_features:
            rating_feats_out = [f for i in range(len(self.rating_generators)) for f in self.rating_generators[i].known_features_return]
            cols_to_add = self.predictor.columns_added + [cn.match_id, cn.team_id, cn.player_id] + rating_feats_out
        else:
            cols_to_add = self.predictor.columns_added + [cn.match_id, cn.team_id, cn.player_id]

        return df.merge(
            df_with_predict[
                cols_to_add
            ],
            on=[cn.match_id, cn.team_id, cn.player_id],
            how="left",
        )

    @property
    def classes_(self) -> Optional[list[str]]:
        """
        Returns the classes of the predictor if available
        """
        if "classes_" not in dir(self.predictor.estimator):
            return None
        return self.predictor.estimator.classes_



import copy
import logging
import warnings
import pandas as pd
from lightgbm import LGBMClassifier, LGBMRegressor
from pandas.errors import SettingWithCopyWarning
from sklearn import clone

from player_performance_ratings.predictor.sklearn_estimator import OrdinalClassifier
from player_performance_ratings.predictor_transformer import PredictorTransformer
from player_performance_ratings.predictor_transformer._simple_transformer import SimpleTransformer
from player_performance_ratings.scorer.score import Filter, apply_filters, Operator

warnings.simplefilter(action="ignore", category=SettingWithCopyWarning)

from typing import Optional

from sklearn.linear_model import LogisticRegression

from player_performance_ratings.consts import PredictColumnNames

from player_performance_ratings.data_structures import ColumnNames
from player_performance_ratings.predictor._base import BasePredictor


class GameTeamPredictor(BasePredictor):
    """
    Wrapper for sklearn models that predicts game results.
    The GameTeam Predictor is intended to transform predictions from a lower granularity into a GameTeam level.
    If input data is at game-player, data is converted to game_team before being trained
    Similar concept if it is at a granularity below game level.

    Can be used similarly to an Sklearn pipeline by injecting pre_transformers into it.
    By default the Predictor will always create pre_transformers to ensure that the estimator can train on the estimator-features that it receives.
    Adding basic encoding of categorical features, standardizing or imputation is therefore not required.

    """

    def __init__(
        self,
        game_id_colum: str,
        team_id_column: str,
        target: Optional[str] = PredictColumnNames.TARGET,
        estimator: Optional = None,
        estimator_features: Optional[list[str]] = None,
        multiclassifier: bool = False,
        pred_column: Optional[str] = None,
        pre_transformers: Optional[list[PredictorTransformer]] = None,
        post_predict_transformers: Optional[list[SimpleTransformer]] = None,
        filters: Optional[list[Filter]] = None,
        multiclass_output_as_struct: bool = False
    ):
        """
        :param game_id_colum - name of game_id column
        :param team_id_column - name of team_id column
        :param target - Name of the column that the predictor should predict
        :param estimator: Sklearn like Estimator
        :param estimator_features: Features that the estimator should use to train.
            Note the estimator_features passed to the constructor can be overriden by estimator_features passed to .train()
        :param multiclassifier: If set to true the output when calling add_prediction() will be in multiclassifier format.
            This results in the pred_column containing a list of probabilities along with a class column added containing the unique classes.
            Further a Logistic Regression estimator will be converted to OrdinalClassifier(LogisticRegression())

        :param pred_column: Name of the new column added containing predictions or probabilities when calling .add_prediction().
            Defaults to f"{self._target}_prediction"
        :param pre_transformers - Transformations to take place before interacting with the estimator.
            The effect is that each Predictor grants the same functionality as an Sklearn Pipeline.
            By default the Predictor will always create pre_transformers to ensure that the estimator can train on the estimator-features that it receives.
            Adding basic encoding of categorical features, standardizing or imputation is therefore not required.
        :param filters - If filters are added the predictor will only train on a subset of the data.
        """

        self.game_id_colum = game_id_colum
        self.team_id_column = team_id_column
        self._target = target
        self._estimator_features = []

        self.multiclassifier = multiclassifier
        super().__init__(
            multiclass_output_as_struct=multiclass_output_as_struct,
            target=self._target,
            pred_column=pred_column,
            estimator=estimator or LogisticRegression(),
            pre_transformers=pre_transformers,
            estimator_features=estimator_features,
            filters=filters,
            post_predict_transformers=post_predict_transformers
        )

    def train(
        self, df: pd.DataFrame, estimator_features: list[Optional[str]] = None
    ) -> None:
        """
        Performs pre_transformations and trains an Sklearn-like estimator.

        :param df - Dataframe containing the estimator_features and target.
        :param estimator_features - If Estimator features are passed they will the estimator_features created by the constructor
        """

        if len(df) == 0:
            raise ValueError("df is empty")

        if estimator_features is None and self._estimator_features is None:
            raise ValueError(
                "estimator features must either be passed to .train() or injected into constructor"
            )

        self._estimator_features = estimator_features or self._estimator_features
        self._estimator_features = self._estimator_features.copy()
        df = apply_filters(df=df, filters=self.filters)
        df = self.fit_transform_pre_transformers(df=df)
        if len(df[self._target].unique()) > 2 and hasattr(
            self.estimator, "predict_proba"
        ):
            self.multiclassifier = True
            if self.estimator.__class__.__name__ == "LogisticRegression":
                self.estimator = OrdinalClassifier(self.estimator)

        if hasattr(self.estimator, "predict_proba"):
            try:
                df = df.assign(**{self._target: df[self._target].astype("int")})
            except Exception:
                pass

        if self._target not in df.columns:
            raise ValueError(f"target {self._target} not in df")

        grouped = self._create_grouped(df)
        self.estimator.fit(grouped[self._estimator_features], grouped[self._target])

    def add_prediction(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Adds prediction to df

        :param df:
        :return: Input df with prediction column
        """

        if hasattr(self.estimator, "predict_proba"):
            try:
                df = df.assign(**{self._target: df[self._target].astype("int")})
            except Exception:
                pass
        if not self._estimator_features:
            raise ValueError("estimator_features not set. Please train first")
        df = self.transform_pre_transformers(df=df)
        grouped = self._create_grouped(df)

        if self.multiclassifier:
            grouped[self._pred_column] = self.estimator.predict_proba(
                grouped[self._estimator_features]
            ).tolist()
            grouped["classes"] = [
                list(self.estimator.classes_) for _ in range(len(grouped))
            ]
        elif not hasattr(self.estimator, "predict_proba"):
            grouped[self._pred_column] = self.estimator.predict(
                grouped[self._estimator_features]
            )
        else:
            grouped[self._pred_column] = self.estimator.predict_proba(
                grouped[self._estimator_features]
            )[:, 1]

        if self.pred_column in df.columns:
            df = df.drop(columns=[self.pred_column])

        if "classes" in grouped.columns:
            df = df.merge(
                grouped[
                    [self.game_id_colum, self.team_id_column]
                    + [self._pred_column, "classes"]
                ],
                on=[self.game_id_colum, self.team_id_column],
            )

        else:
            df = df.merge(
                grouped[
                    [self.game_id_colum, self.team_id_column] + [self._pred_column]
                ],
                on=[self.game_id_colum, self.team_id_column],
            )

        for simple_transformer in self.post_predict_transformers:
            df = simple_transformer.transform(df)

        if self.multiclass_output_as_struct  and self.multiclassifier:
            df = self._convert_multiclass_predictions_to_struct(df)

        return df

    def _create_grouped(self, df: pd.DataFrame) -> pd.DataFrame:

        numeric_features = [
            feature
            for feature in self._estimator_features
            if df[feature].dtype in ["int", "float"]
        ]
        cat_feats = [
            feature
            for feature in self._estimator_features
            if feature not in numeric_features
        ]

        if self._target in df.columns:
            if df[self._target].dtype == "object":
                df.loc[:, self._target] = df[self._target].astype("int")

            grouped = (
                df.groupby([self.game_id_colum, self.team_id_column])
                .agg(
                    {
                        **{feature: "mean" for feature in numeric_features},
                        self._target: "mean",
                    }
                )
                .reset_index()
            )

        else:
            grouped = (
                df.groupby([self.game_id_colum, self.team_id_column])
                .agg({**{feature: "mean" for feature in numeric_features}})
                .reset_index()
            )

        if self._target in df.columns and hasattr(
            self._deepest_estimator, "predict_proba"
        ):
            grouped[self._target] = grouped[self._target].astype("int")

        grouped = grouped.merge(
            df[[self.game_id_colum, self.team_id_column, *cat_feats]].drop_duplicates(
                subset=[self.game_id_colum, self.team_id_column]
            ),
            on=[self.game_id_colum, self.team_id_column],
            how="inner",
        )

        return grouped


class Predictor(BasePredictor):
    """
    Wrapper for sklearn models that predicts game results.
    Can be used similarly to an Sklearn pipeline by injecting pre_transformers into it.
    By default the Predictor will always create pre_transformers to ensure that the estimator can train on the estimator-features that it receives.
    Adding basic encoding of categorical features, standardizing or imputation is therefore not required.
    """

    def __init__(
        self,
        target: Optional[str] = PredictColumnNames.TARGET,
        estimator: Optional = None,
        estimator_features: Optional[list[str]] = None,
        filters: Optional[list[Filter]] = None,
        multiclassifier: bool = False,
        pred_column: Optional[str] = None,
        column_names: Optional[ColumnNames] = None,
        pre_transformers: Optional[list[PredictorTransformer]] = None,
        post_predict_transformers: Optional[list[SimpleTransformer]] = None,
        multiclass_output_as_struct: bool = False,
    ):
        """
        :param target - Name of the column that the predictor should predict
        :param estimator: Sklearn like Estimator
        :param estimator_features: Features that the estimator should use to train.
            Note the estimator_features passed to the constructor can be overriden by estimator_features passed to .train()
        :param multiclassifier: If set to true the output when calling add_prediction() will be in multiclassifier format.
            This results in the pred_column containing a list of probabilities along with a class column added containing the unique classes.
            Further a Logistic Regression estimator will be converted to OrdinalClassifier(LogisticRegression())
        :param pred_column: Name of the new column added containing predictions or probabilities when calling .add_prediction().
            Defaults to f"{self._target}_prediction"
        :param pre_transformers - Transformations to take place before interacting with the estimator.
            The effect is that each Predictor grants the same functionality as an Sklearn Pipeline.
            By default the Predictor will always create pre_transformers to ensure that the estimator can train on the estimator-features that it receives.
            Adding basic encoding of categorical features, standardizing or imputation is therefore not required.
        :param filters - If filters are added the predictor will only train on a subset of the data.
        """
        self._target = target
        self.multiclassifier = multiclassifier
        self.column_names = column_names

        if estimator is None:
            logging.warning(
                "model is not set. Will use LGBMClassifier(max_depth=2, n_estimators=100, learning_rate=0.1)"
            )

        super().__init__(
            target=self._target,
            multiclass_output_as_struct=multiclass_output_as_struct,
            pred_column=pred_column,
            estimator=estimator
            or LGBMClassifier(max_depth=2, n_estimators=100, verbose=-100),
            pre_transformers=pre_transformers,
            post_predict_transformers=post_predict_transformers,
            filters=filters,
            estimator_features=estimator_features,
        )

    def train(
        self, df: pd.DataFrame, estimator_features: Optional[list[str]] = None
    ) -> None:
        """
        Performs pre_transformations and trains an Sklearn-like estimator.

        :param df - Dataframe containing the estimator_features and target.
        :param estimator_features - If Estimator features are passed they will the estimator_features created by the constructor
        """

        if len(df) == 0:
            raise ValueError("df is empty")

        if estimator_features is None and self._estimator_features is None:
            raise ValueError(
                "estimator features must either be passed to .train() or injected into constructor"
            )
        self._estimator_features = estimator_features or self._estimator_features
        self._estimator_features = self._estimator_features.copy()

        filtered_df = apply_filters(df=df, filters=self.filters)
        if hasattr(self.estimator, "predict_proba"):
            try:
                filtered_df[self._target] = filtered_df[self._target].astype("int")
            except Exception:
                pass

        filtered_df = self.fit_transform_pre_transformers(df=filtered_df)

        if (
            not self.multiclassifier
            and len(filtered_df[self._target].unique()) > 2
            and hasattr(self._deepest_estimator, "predict_proba")
        ):
            self.multiclassifier = True
            if self.estimator.__class__.__name__ == "LogisticRegression":
                self.estimator = OrdinalClassifier(self.estimator)
            if len(filtered_df[self._target].unique()) > 50:
                logging.warning(
                    f"target has {len(df[self._target].unique())} unique values. This may machine-learning model to not function properly."
                    f" It is recommended to limit max and min values to ensure less than 50 unique targets"
                )

        if hasattr(self._deepest_estimator, "predict_proba"):
            filtered_df = filtered_df.assign(
                **{self._target: filtered_df[self._target].astype("int")}
            )

        self.estimator.fit(
            filtered_df[self._estimator_features], filtered_df[self._target]
        )

    def add_prediction(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Adds prediction to df

        :param df:
        :return: Input df with prediction column
        """
        df = df.copy()
        if not self._estimator_features:
            raise ValueError("estimator_features not set. Please train first")

        if hasattr(self.estimator, "predict_proba"):
            try:
                df[self._target] = df[self._target].astype("int")
            except Exception:
                pass

        df = self.transform_pre_transformers(df=df)
        if self.multiclassifier:
            df[self._pred_column] = self.estimator.predict_proba(
                df[self._estimator_features]
            ).tolist()
            df["classes"] = [list(self.estimator.classes_) for _ in range(len(df))]
            if len(set(df[self.pred_column].iloc[0])) == 2:
                raise ValueError(
                    "Too many unique values in relation to rows in the training dataset causes multiclassifier to not train properly"
                )

        elif not hasattr(self._deepest_estimator, "predict_proba"):
            df[self._pred_column] = self.estimator.predict(df[self._estimator_features])
        else:
            df[self._pred_column] = self.estimator.predict_proba(
                df[self._estimator_features]
            )[:, 1]

        for simple_transformer in self.post_predict_transformers:
            df = simple_transformer.transform(df)

        if self.multiclass_output_as_struct and self.multiclassifier:
            df = self._convert_multiclass_predictions_to_struct(df)

        return df


class GranularityPredictor(BasePredictor):
    """
    Samples the dataset into different subsets based on the granularity column and trains a separate estimator for each.
    """

    def __init__(
        self,
        granularity_column_name: str,
        target: Optional[str] = PredictColumnNames.TARGET,
        estimator: Optional = None,
        estimator_features: Optional[list[str]] = None,
        filters: Optional[list[Filter]] = None,
        multiclassifier: bool = False,
        pred_column: Optional[str] = None,
        column_names: Optional[ColumnNames] = None,
        pre_transformers: Optional[list[PredictorTransformer]] = None,
        multiclass_output_as_struct: bool =False,
    ):
        """
        :param target - Name of the column that the predictor should predict
        :param estimator: Sklearn like Estimator
        :param estimator_features: Features that the estimator should use to train.
            Note the estimator_features passed to the constructor can be overriden by estimator_features passed to .train()
        :param multiclassifier: If set to true the output when calling add_prediction() will be in multiclassifier format.
            This results in the pred_column containing a list of probabilities along with a class column added containing the unique classes.
            Further a Logistic Regression estimator will be converted to OrdinalClassifier(LogisticRegression())
        :param pred_column: Name of the new column added containing predictions or probabilities when calling .add_prediction().
            Defaults to f"{self._target}_prediction"
        :param pre_transformers - Transformations to take place before interacting with the estimator.
            The effect is that each Predictor grants the same functionality as an Sklearn Pipeline.
            By default the Predictor will always create pre_transformers to ensure that the estimator can train on the estimator-features that it receives.
            Adding basic encoding of categorical features, standardizing or imputation is therefore not required.
        :param filters - If filters are added the predictor will only train on a subset of the data.
        """

        self._target = target
        self.granularity_column_name = granularity_column_name
        self.multiclassifier = multiclassifier
        self.column_names = column_names
        self._granularities = []
        self._granularity_estimators = {}

        if estimator is None:
            logging.warning(
                "model is not set. Will use LGBMClassifier(max_depth=2, n_estimators=100)"
            )

        super().__init__(
            target=self._target,
            pred_column=pred_column,
            estimator=estimator
            or LGBMClassifier(max_depth=2, n_estimators=100, verbose=-100),
            pre_transformers=pre_transformers,
            filters=filters,
            estimator_features=estimator_features,
            multiclass_output_as_struct=multiclass_output_as_struct,
            post_predict_transformers=[]
        )

    def train(self, df: pd.DataFrame, estimator_features: list[str]) -> None:
        """
        Performs pre_transformations and trains an Sklearn-like estimator.

        :param df - Dataframe containing the estimator_features and target.
        :param estimator_features - If Estimator features are passed they will the estimator_features created by the constructor
        """

        if len(df) == 0:
            raise ValueError("df is empty")

        if estimator_features is None and self._estimator_features is None:
            raise ValueError(
                "estimator features must either be passed to .train() or injected into constructor"
            )
        self._estimator_features = estimator_features or self._estimator_features
        self._estimator_features = self._estimator_features.copy()

        filtered_df = apply_filters(df=df, filters=self.filters)
        if hasattr(self.estimator, "predict_proba"):
            try:
                filtered_df[self._target] = filtered_df[self._target].astype("int")
            except Exception:
                pass

        filtered_df = self.fit_transform_pre_transformers(df=filtered_df)

        if hasattr(self._deepest_estimator, "predict_proba"):
            filtered_df = filtered_df.assign(
                **{self._target: filtered_df[self._target].astype("int")}
            )

        if (
            not self.multiclassifier
            and len(filtered_df[self._target].unique()) > 2
            and hasattr(self._deepest_estimator, "predict_proba")
        ):
            self.multiclassifier = True
            if self.estimator.__class__.__name__ == "LogisticRegression":
                self.estimator = OrdinalClassifier(self.estimator)
            if len(filtered_df[self._target].unique()) > 50:
                logging.warning(
                    f"target has {len(df[self._target].unique())} unique values. This may machine-learning model to not function properly."
                    f" It is recommended to limit max and min values to ensure less than 50 unique targets"
                )

        self._granularities = filtered_df[self.granularity_column_name].unique()

        for granularity in self._granularities:
            self._granularity_estimators[granularity] = clone(self.estimator)
            rows = filtered_df[filtered_df[self.granularity_column_name] == granularity]
            self._granularity_estimators[granularity].fit(
                rows[self._estimator_features], rows[self._target]
            )

    def add_prediction(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Adds prediction to df

               :param df:
               :return: Input df with prediction column
        """
        if not self._estimator_features:
            raise ValueError("estimator_features not set. Please train first")

        if hasattr(self.estimator, "predict_proba"):
            try:
                df = df.assign(**{self._target: df[self._target].astype("int")})
            except Exception:
                pass

        df = self.transform_pre_transformers(df=df)
        dfs = []
        for granularity, estimator in self._granularity_estimators.items():
            rows = df[df[self.granularity_column_name] == granularity]
            if self.multiclassifier:
                rows = rows.assign(
                    **{
                        self._pred_column: estimator.predict_proba(
                            rows[self._estimator_features]
                        ).tolist()
                    }
                )
                rows = rows.assign(
                    **{"classes": [list(estimator.classes_) for _ in range(len(rows))]}
                )
                if len(set(rows[self.pred_column].iloc[0])) == 2:
                    raise ValueError(
                        "Too many unique values in relation to rows in the training dataset causes multiclassifier to not train properly"
                    )

            elif not hasattr(self._deepest_estimator, "predict_proba"):
                rows = rows.assign(
                    **{
                        self._pred_column: estimator.predict(
                            rows[self._estimator_features]
                        )
                    }
                )
            else:
                rows = rows.assign(
                    **{
                        self._pred_column: estimator.predict_proba(
                            rows[self._estimator_features]
                        )[:, 1]
                    }
                )
            dfs.append(rows)

        df = pd.concat(dfs)
        for simple_transformer in self.post_predict_transformers:
            df = simple_transformer.transform(df)
        return df


class SeriesWinLosePredictor(BasePredictor):
    """
    Trains a separate model for when target is 1 vs when target is 0.
    For a bo1, the probability is then calculated as Probabilities given team wins * Game Win Probability + (1-Game Win Probabiliy) * Probability Given Team Loses
    This can be extended to bo3, bo5 based on the format-column.
    """

    def __init__(
        self,
        format_column_name: str,
        game_win_prob_column_name: str,
        target: Optional[str] = PredictColumnNames.TARGET,
        estimator: Optional = None,
        estimator_features: Optional[list[str]] = None,
        filters: Optional[list[Filter]] = None,
        multiclassifier: bool = False,
        pred_column: Optional[str] = None,
        column_names: Optional[ColumnNames] = None,
        pre_transformers: Optional[list[PredictorTransformer]] = None,
    ):
        self._target = target
        self.format_column_name = format_column_name
        self.game_win_prob_column_name = game_win_prob_column_name
        self.multiclassifier = multiclassifier
        self.column_names = column_names
        win_filters = filters.copy()
        win_filters.append(
            Filter(
                column_name=self.game_win_prob_column_name,
                value=1,
                operator=Operator.EQUALS,
            )
        )
        lose_filters = filters.copy()
        lose_filters.append(
            Filter(
                column_name=self.game_win_prob_column_name,
                value=0,
                operator=Operator.EQUALS,
            )
        )

        self._win_predictor = Predictor(
            target=self._target,
            estimator=estimator,
            estimator_features=estimator_features,
            filters=win_filters,
            multiclassifier=multiclassifier,
            pred_column=pred_column,
            column_names=column_names,
            pre_transformers=pre_transformers,
        )
        self._lose_predictor = Predictor(
            target=self._target,
            estimator=estimator,
            estimator_features=estimator_features,
            filters=lose_filters,
            multiclassifier=multiclassifier,
            pred_column=pred_column,
            column_names=column_names,
            pre_transformers=pre_transformers,
        )

        if estimator is None:
            logging.warning(
                "model is not set. Will use LGBMClassifier(max_depth=2, n_estimators=100)"
            )

        super().__init__(
            target=self._target,
            pred_column=pred_column,
            estimator=estimator
            or LGBMClassifier(max_depth=2, n_estimators=100, verbose=-100),
            pre_transformers=pre_transformers,
            filters=filters,
            estimator_features=estimator_features,
        )

    def train(self, df: pd.DataFrame, estimator_features: list[str]) -> None:
        self._win_predictor.train(df=df, estimator_features=estimator_features)
        self._lose_predictor.train(df=df, estimator_features=estimator_features)

    def add_prediction(self, df: pd.DataFrame) -> pd.DataFrame:
        win_df = df[df[self.game_win_prob_column_name] == 1]
        lose_df = df[df[self.game_win_prob_column_name] == 0]

        win_df = self._win_predictor.add_prediction(win_df)
        lose_df = self._lose_predictor.add_prediction(lose_df)

        return pd.concat([win_df, lose_df]).sort_values(
            by=[
                self.column_names.start_date,
                self.column_names.match_id,
                self.column_names.team_id,
                self.column_names.player_id,
            ]
        )


class PointToClassificationPredictor(BasePredictor):

    def __init__(
        self,
        target: Optional[str] = PredictColumnNames.TARGET,
        estimator: Optional = None,
        point_estimate_column: Optional[str] = None,
        estimator_features: Optional[list[str]] = None,
        filters: Optional[list[Filter]] = None,
        multiclassifier: bool = False,
        pred_column: Optional[str] = None,
        column_names: Optional[ColumnNames] = None,
        pre_transformers: Optional[list[PredictorTransformer]] = None,
    ):
        self._target = target
        self.multiclassifier = multiclassifier
        self.point_estimate_column = point_estimate_column
        self.column_names = column_names
        self._target_probs = {}
        super().__init__(
            target=self._target,
            pred_column=pred_column,
            estimator=estimator
            or LGBMRegressor(
                max_depth=2, n_estimators=100, learning_rate=0.05, verbose=-100
            ),
            pre_transformers=pre_transformers,
            filters=filters,
            estimator_features=estimator_features,
        )

    def train(self, df: pd.DataFrame, estimator_features: list[str]) -> None:
        if self.point_estimate_column is not None:
            predictions = df[self.point_estimate_column]
        else:
            self.estimator.fit(df[estimator_features], df[self._target])
            predictions = self.estimator.predict(df[estimator_features])

        quantiles = predictions.quantile([q / 50 for q in range(1, 50)])

        for idx, quantile in enumerate(quantiles[:-1]):
            rows = df[(predictions >= quantile) & (predictions < quantiles[idx + 1])]
            value_counts = rows[self._target].value_counts()
            for target, count in value_counts:
                self._target_probs[target] = count / len(df)



import logging
from dataclasses import dataclass
from enum import Enum
from typing import Optional

import numpy as np
import pandas as pd
from player_performance_ratings.predictor import BasePredictor

from player_performance_ratings import ColumnNames

from player_performance_ratings.transformers.base_transformer import (
    BaseTransformer,
    BaseLagGenerator,
)
import narwhals as nw
from narwhals.typing import FrameT


class NetOverPredictedPostTransformer(BaseTransformer):

    def __init__(
            self,
            predictor: BasePredictor,
            features: list[str] = None,
            lag_generators: Optional[list[BaseLagGenerator]] = None,
            prefix: str = "net_over_predicted_",
            are_estimator_features: bool = False,
    ):
        super().__init__(
            features=features,
            are_estimator_features=are_estimator_features,
            features_out=[],
        )
        self.prefix = prefix
        self.predictor = predictor
        self._features_out = []
        self.lag_generators = lag_generators or []
        self.column_names = None
        new_feature_name = self.prefix + self.predictor.pred_column
        self._features_out.append(new_feature_name)
        self._estimator_features_out = []
        for lag_generator in self.lag_generators:
            if not lag_generator.features:
                lag_generator.features = [new_feature_name]
                for iteration in lag_generator.iterations:
                    lag_generator._features_out = [
                        f"{lag_generator.prefix}{iteration}_{new_feature_name}"
                    ]
                    self.features_out.extend(lag_generator._features_out.copy())
                    self._estimator_features_out.extend(
                        lag_generator._features_out.copy()
                    )

        if self._are_estimator_features:
            self._estimator_features_out.append(self.predictor.pred_column)
            self.features_out.append(self.predictor.pred_column)
        if self.prefix is "":
            raise ValueError("Prefix must not be empty")

    def fit_transform(
            self, df: pd.DataFrame, column_names: Optional[ColumnNames] = None
    ) -> pd.DataFrame:
        ori_cols = df.columns.tolist()
        self.column_names = column_names
        self.predictor.train(df, estimator_features=self.features)
        df = self.predictor.add_prediction(df)
        new_feature_name = self.prefix + self.predictor.pred_column
        if self.predictor.target not in df.columns:
            df = df.assign(**{new_feature_name: np.nan})
        else:
            df = df.assign(
                **{
                    new_feature_name: df[self.predictor.target]
                                      - df[self.predictor.pred_column]
                }
            )
        for lag_generator in self.lag_generators:
            df = lag_generator.generate_historical(df, column_names=self.column_names)
        return df[list(set(ori_cols + self.features_out))]

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        ori_cols = df.columns.tolist()
        df = self.predictor.add_prediction(df)
        new_feature_name = self.prefix + self.predictor.pred_column
        if self.predictor.target not in df.columns:
            df = df.assign(**{new_feature_name: np.nan})
        else:
            df = df.assign(
                **{
                    new_feature_name: df[self.predictor.target]
                                      - df[self.predictor.pred_column]
                }
            )

        for lag_generator in self.lag_generators:
            df = lag_generator.generate_future(df)

        return df[list(set(ori_cols + self.features_out))]

    @property
    def features_out(self) -> list[str]:
        return self._features_out

    def reset(self) -> "BaseTransformer":
        for lag_generator in self.lag_generators:
            lag_generator.reset()
        return self


class Operation(Enum):
    SUBTRACT = "subtract"
    MULTIPLY = "multiply"


@dataclass
class ModifyOperation:
    feature1: str
    operation: Operation
    feature2: str
    new_column_name: Optional[str] = None

    def __post_init__(self):
        if self.operation == Operation.SUBTRACT and not self.new_column_name:
            self.new_column_name = f"{self.feature1}_minus_{self.feature2}"


class OperatorTransformer(BaseTransformer):
    """
    Performs operations on two columns and stores the result in a new column.
    An operation can be subtraction, addition, multiplication, or division.
    """

    def __init__(
            self,
            feature1: str,
            operation: Operation,
            feature2: str,
            new_column_name: Optional[str] = None,
            features: Optional[list[str]] = None,
            are_estimator_features: bool = True,
    ):
        """
        :param feature1: The first feature to perform the operation on
        :param operation: The operation to perform
        :param feature2: The second feature to perform the operation on
        :param new_column_name: The name of the new column to store the result in
        """
        self.feature1 = feature1
        self.operation = operation
        self.feature2 = feature2
        self.new_column_name = new_column_name

        if not self.new_column_name:

            if self.operation == Operation.SUBTRACT:
                self.new_column_name = f"{self.feature1}_minus_{self.feature2}"
            elif self.operation == Operation.MULTIPLY:
                self.new_column_name = f"{self.feature1}_times_{self.feature2}"

        super().__init__(
            features=features,
            are_estimator_features=are_estimator_features,
             features_out=[self.new_column_name],
        )

    def fit_transform(
            self, df: FrameT, column_names: Optional[ColumnNames]
    ) -> FrameT:
        self.column_names = column_names
        return self.transform(df)

    @nw.narwhalify
    def transform(self, df: FrameT) -> FrameT:
        if self.operation == Operation.SUBTRACT:
            df = df.with_columns(
                (nw.col(self.feature1) - nw.col(self.feature2))
                .alias(self.new_column_name)
            )

        elif self.operation == Operation.MULTIPLY:
            df = df.with_columns(
                (nw.col(self.feature1) * nw.col(self.feature2))
                .alias(self.new_column_name)
            )
        else:
            logging.warning(f"Operation {self.operation} not implemented")
            raise NotImplementedError

        return df


class ModifierTransformer(BaseTransformer):
    """
    Performs operations on two columns and stores the result in a new column.
    An operation can be subtraction, addition, multiplication, or division.
    """

    def __init__(
            self,
            modify_operations: list[ModifyOperation],
            features: Optional[list[str]] = None,
            are_estimator_features: bool = True,
    ):
        """
        :param modify_operations: A list of ModifyOperations to perform
        :param features: Features need to be passed due to BaseTransformer requring them, although has no effect on the transformer
            Thus keep it as default None.
        :param are_estimator_features: Whether the new created columns should be used as estimator features
        """
        self.modify_operations = modify_operations
        _features_out = [
            operation.new_column_name for operation in self.modify_operations
        ]
        super().__init__(
            features=features,
            are_estimator_features=are_estimator_features,
            features_out=_features_out,
        )

    def fit_transform(
            self, df: pd.DataFrame, column_names: Optional[ColumnNames]
    ) -> pd.DataFrame:
        self.column_names = column_names
        return self.transform(df)

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        for operation in self.modify_operations:
            if operation.operation == Operation.SUBTRACT:
                if (
                        operation.feature1 not in df.columns
                        or operation.feature2 not in df.columns
                ):
                    df = df.assign(**{operation.new_column_name: np.nan})

                else:
                    df = df.assign(
                        **{
                            operation.new_column_name: df[operation.feature1]
                                                       - df[operation.feature2]
                        }
                    )

        return df


class PredictorTransformer(BaseTransformer):
    """
    Transformer that uses a predictor to generate predictions on the dataset
    This is useful if you want to use the output of a feature as input for another model
    """

    def __init__(self, predictor: BasePredictor, features: list[str] = None):
        """
        :param predictor: The predictor to use to add add new prediction-columns to the dataset
        :param features: The features to use for the predictor
        """
        self.predictor = predictor
        super().__init__(
            features=features, features_out=[f"{self.predictor.pred_column}"]
        )

    def fit_transform(
            self, df: pd.DataFrame, column_names: Optional[None] = None
    ) -> pd.DataFrame:
        self.predictor.train(df=df, estimator_features=self.features)
        return self.transform(df)

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        df = self.predictor.add_prediction(df=df)
        return df


class RatioTeamPredictorTransformer(BaseTransformer):
    """
    Transformer that trains and uses the output of a predictor and divides it by the sum of the predictions for all the players within the team
    If team_total_prediction_column is passed in, it will also multiply the ratio by the team_total_prediction_column
    This is useful to provide a normalized point-estimate for a player for the given feature.

    """

    def __init__(
            self,
            features: list[str],
            predictor: BasePredictor,
            team_total_prediction_column: Optional[str] = None,
            lag_generators: Optional[list[BaseLagGenerator]] = None,
            prefix: str = "_ratio_team",
    ):
        """
        :param features: The features to use for the predictor
        :param predictor: The predictor to use to add add new prediction-columns to the dataset
        :param team_total_prediction_column: If passed, The column to multiply the ratio by.
        :param lag_generators: Additional lag-generators (such as rolling-mean) can be performed after the ratio is calculated is passed
        :param prefix: The prefix to use for the new columns
        """

        self.predictor = predictor
        self.team_total_prediction_column = team_total_prediction_column
        self.prefix = prefix
        self.predictor._pred_column = f"__prediction__{self.predictor.target}"
        self.lag_generators = lag_generators or []
        super().__init__(
            features=features,
            features_out=[self.predictor.target + prefix, self.predictor._pred_column],
        )

        if self.team_total_prediction_column:
            self._features_out.append(
                self.predictor.target + prefix + "_team_total_multiplied"
            )
        for lag_generator in self.lag_generators:
            self._features_out.extend(lag_generator.features_out)

        if self._are_estimator_features:
            self._estimator_features_out = self._features_out.copy()

    def fit_transform(
            self, df: pd.DataFrame, column_names: Optional[ColumnNames] = None
    ) -> pd.DataFrame:
        ori_cols = df.columns.tolist()
        self.column_names = column_names
        self.predictor.train(df=df, estimator_features=self.features)
        transformed_df = self.transform(df)
        for lag_generator in self.lag_generators:
            transformed_df = lag_generator.generate_historical(
                transformed_df, column_names=self.column_names
            )

        return transformed_df[list(set(ori_cols + self.features_out))]

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        df = self.predictor.add_prediction(df=df)

        df[self.predictor.pred_column + "_sum"] = df.groupby(
            [self.column_names.match_id, self.column_names.team_id]
        )[self.predictor.pred_column].transform("sum")
        df[self._features_out[0]] = (
                df[self.predictor.pred_column] / df[self.predictor.pred_column + "_sum"]
        )
        if self.team_total_prediction_column:
            df = df.assign(
                **{
                    self.predictor.target
                    + self.prefix
                    + "_team_total_multiplied": df[self._features_out[0]]
                                                * df[self.team_total_prediction_column]
                }
            )

        for lag_transformer in self.lag_generators:
            df = lag_transformer.generate_historical(df, column_names=self.column_names)

        return df.drop(columns=[self.predictor.pred_column + "_sum"])

    @property
    def features_out(self) -> list[str]:
        return self._features_out

    def reset(self) -> "BaseTransformer":
        for lag_generator in self.lag_generators:
            lag_generator.reset()
        return self



import pandas as pd
import pytest
import polars as pl

from player_performance_ratings import ColumnNames
from player_performance_ratings.pipeline_transformer import PipelineTransformer
from player_performance_ratings.ratings import (
    UpdateRatingGenerator,
    RatingKnownFeatures,
)
from player_performance_ratings.ratings.performance_generator import (
    ColumnWeight,
    Performance,
    PerformancesGenerator,
)
from player_performance_ratings.transformers import LagTransformer


@pytest.mark.parametrize("to_polars", [True, False])
def test_pipeline_transformer(to_polars):
    df = pd.DataFrame(
        {
            "game_id": [1, 1, 2, 2, 3, 3, 4, 4, 5, 5],
            "player_id": [1, 2, 1, 2, 1, 2, 1, 2, 1, 2],
            "team_id": [1, 2, 1, 2, 1, 2, 1, 2, 1, 2],
            "start_date": [
                pd.to_datetime("2023-01-01"),
                pd.to_datetime("2023-01-01"),
                pd.to_datetime("2023-01-02"),
                pd.to_datetime("2023-01-02"),
                pd.to_datetime("2023-01-03"),
                pd.to_datetime("2023-01-03"),
                pd.to_datetime("2023-01-03"),
                pd.to_datetime("2023-01-03"),
                pd.to_datetime("2023-01-04"),
                pd.to_datetime("2023-01-04"),
            ],
            "deaths": [0.5, 1, 0.7, 2, 0.5, 0.7, 0.2, 2.1, 0.8, 1],
            "kills": [0.2, 0.3, 0.2, 0.3, 0.2, 0.3, 0.2, 0.3, 0.2, 0.3],
            "__target": [1, 0, 0.6, 0.3, 0.8, 0.2, 0.4, 0.1, 1, 0],
        }
    )

    column_weights = [ColumnWeight(name="kills", weight=1)]
    column_names = ColumnNames(
        match_id="game_id",
        team_id="team_id",
        player_id="player_id",
        start_date="start_date",
    )
    rating_generator = UpdateRatingGenerator(
        known_features_out=[RatingKnownFeatures.RATING_DIFFERENCE_PROJECTED]
    )

    pipeline = PipelineTransformer(
        performances_generator=PerformancesGenerator(
            Performance(weights=column_weights)
        ),
        rating_generators=rating_generator,
        lag_generators=[
            LagTransformer(
                features=["kills", "deaths"], lag_length=1, granularity=["player_id"]
            )
        ],
        column_names=column_names,
    )


    historical_df = df[df[column_names.start_date] < pd.to_datetime("2023-01-04")]
    future_df = df[df[column_names.start_date] >= pd.to_datetime("2023-01-04")]

    if to_polars:
        historical_df = pl.DataFrame(df)
        future_df = pl.DataFrame(df)

    historical_df_predictions = pipeline.fit_transform(df=historical_df)
    future_predict = pipeline.transform(df=future_df)
    if to_polars:
        historical_df_predictions = historical_df_predictions.to_pandas()
        future_predict = future_predict.to_pandas()

    for lag_generator in pipeline.lag_generators:
        for feature_out in lag_generator.features_out:
            assert feature_out in historical_df_predictions.columns
            assert feature_out in future_predict.columns

    for rating_feature in rating_generator.features_out:
        assert rating_feature in historical_df_predictions.columns
        assert rating_feature in future_predict.columns

    for post_lag_transformer in pipeline.post_lag_transformers:
        assert post_lag_transformer in historical_df_predictions.columns
        assert post_lag_transformer in future_predict.columns
